baseVersion = '8.2.0'
cgcloudVersion = '1.6.0a1.dev393'
version = '8.2.0-3efbe6f84b89a1eeab1d7b5913bbd01084d6490e'
cacheTag = 'cache-local-py3.10'
mainCacheTag = 'cache-master-py3.10'
distVersion = '8.2.0'
exactPython = 'python3.10'
python = 'python3.10'
dockerTag = '8.2.0-3efbe6f84b89a1eeab1d7b5913bbd01084d6490e-py3.10'
currentCommit = '3efbe6f84b89a1eeab1d7b5913bbd01084d6490e'
dockerRegistry = 'quay.io/ucsc_cgl'
dockerName = 'toil'
dirty = False
cwltool_version = '3.1.20250110105449'
