from textwrap import TextWrapper
from typing import Sequence, Tuple

from .topicexc import TopicNameError

UNDERSCORE: str
ALL_TOPICS: str

class WeakNone:
    def __call__(self) -> None: ...

def smartDedent(paragraph: str) -> str: ...
def validateName(topicName: str) -> None: ...
def stringize(topicName: Sequence[str]) -> str: ...
def tupleize(topicName: str) -> Tuple[str, ...]: ...
