# !!! AUTOGENERATED FILE !!!
# Update with: src/toil/utils/toilUpdateEC2Instances.py
#
# Copyright (C) 2015-2025 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from toil.lib.ec2nodes import InstanceType


# 880 Instance Types.  Generated 2025-02-06 14:13:22.640293.
E2Instances = {
    'a1.2xlarge': InstanceType(name='a1.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'a1.4xlarge': InstanceType(name='a1.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'a1.large': InstanceType(name='a1.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'a1.medium': InstanceType(name='a1.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='amd64'),
    'a1.metal': InstanceType(name='a1.metal', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'a1.xlarge': InstanceType(name='a1.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c1.medium': InstanceType(name='c1.medium', cores=2, memory=1.7, disks=1.0, disk_capacity=350.0, architecture='amd64'),
    'c1.xlarge': InstanceType(name='c1.xlarge', cores=8, memory=7.0, disks=4.0, disk_capacity=420.0, architecture='amd64'),
    'c3.2xlarge': InstanceType(name='c3.2xlarge', cores=8, memory=15.0, disks=2.0, disk_capacity=80.0, architecture='amd64'),
    'c3.4xlarge': InstanceType(name='c3.4xlarge', cores=16, memory=30.0, disks=2.0, disk_capacity=160.0, architecture='amd64'),
    'c3.8xlarge': InstanceType(name='c3.8xlarge', cores=32, memory=60.0, disks=2.0, disk_capacity=320.0, architecture='amd64'),
    'c3.large': InstanceType(name='c3.large', cores=2, memory=3.75, disks=2.0, disk_capacity=16.0, architecture='amd64'),
    'c3.xlarge': InstanceType(name='c3.xlarge', cores=4, memory=7.5, disks=2.0, disk_capacity=40.0, architecture='amd64'),
    'c4.2xlarge': InstanceType(name='c4.2xlarge', cores=8, memory=15.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c4.4xlarge': InstanceType(name='c4.4xlarge', cores=16, memory=30.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c4.8xlarge': InstanceType(name='c4.8xlarge', cores=36, memory=60.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c4.large': InstanceType(name='c4.large', cores=2, memory=3.75, disks=0, disk_capacity=0, architecture='amd64'),
    'c4.xlarge': InstanceType(name='c4.xlarge', cores=4, memory=7.5, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.12xlarge': InstanceType(name='c5.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.18xlarge': InstanceType(name='c5.18xlarge', cores=72, memory=144.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.24xlarge': InstanceType(name='c5.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.2xlarge': InstanceType(name='c5.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.4xlarge': InstanceType(name='c5.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.9xlarge': InstanceType(name='c5.9xlarge', cores=36, memory=72.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.large': InstanceType(name='c5.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.metal': InstanceType(name='c5.metal', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5.xlarge': InstanceType(name='c5.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.12xlarge': InstanceType(name='c5a.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.16xlarge': InstanceType(name='c5a.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.24xlarge': InstanceType(name='c5a.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.2xlarge': InstanceType(name='c5a.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.4xlarge': InstanceType(name='c5a.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.8xlarge': InstanceType(name='c5a.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.large': InstanceType(name='c5a.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5a.xlarge': InstanceType(name='c5a.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5ad.12xlarge': InstanceType(name='c5ad.12xlarge', cores=48, memory=96.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'c5ad.16xlarge': InstanceType(name='c5ad.16xlarge', cores=64, memory=128.0, disks=2.0, disk_capacity=1200.0, architecture='amd64'),
    'c5ad.24xlarge': InstanceType(name='c5ad.24xlarge', cores=96, memory=192.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'c5ad.2xlarge': InstanceType(name='c5ad.2xlarge', cores=8, memory=16.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'c5ad.4xlarge': InstanceType(name='c5ad.4xlarge', cores=16, memory=32.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'c5ad.8xlarge': InstanceType(name='c5ad.8xlarge', cores=32, memory=64.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'c5ad.large': InstanceType(name='c5ad.large', cores=2, memory=4.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'c5ad.xlarge': InstanceType(name='c5ad.xlarge', cores=4, memory=8.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'c5d.12xlarge': InstanceType(name='c5d.12xlarge', cores=48, memory=96.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'c5d.18xlarge': InstanceType(name='c5d.18xlarge', cores=72, memory=144.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'c5d.24xlarge': InstanceType(name='c5d.24xlarge', cores=96, memory=192.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'c5d.2xlarge': InstanceType(name='c5d.2xlarge', cores=8, memory=16.0, disks=1.0, disk_capacity=200.0, architecture='amd64'),
    'c5d.4xlarge': InstanceType(name='c5d.4xlarge', cores=16, memory=32.0, disks=1.0, disk_capacity=400.0, architecture='amd64'),
    'c5d.9xlarge': InstanceType(name='c5d.9xlarge', cores=36, memory=72.0, disks=1.0, disk_capacity=900.0, architecture='amd64'),
    'c5d.large': InstanceType(name='c5d.large', cores=2, memory=4.0, disks=1.0, disk_capacity=50.0, architecture='amd64'),
    'c5d.metal': InstanceType(name='c5d.metal', cores=96, memory=192.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'c5d.xlarge': InstanceType(name='c5d.xlarge', cores=4, memory=8.0, disks=1.0, disk_capacity=100.0, architecture='amd64'),
    'c5n.18xlarge': InstanceType(name='c5n.18xlarge', cores=72, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5n.2xlarge': InstanceType(name='c5n.2xlarge', cores=8, memory=21.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5n.4xlarge': InstanceType(name='c5n.4xlarge', cores=16, memory=42.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5n.9xlarge': InstanceType(name='c5n.9xlarge', cores=36, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5n.large': InstanceType(name='c5n.large', cores=2, memory=5.25, disks=0, disk_capacity=0, architecture='amd64'),
    'c5n.metal': InstanceType(name='c5n.metal', cores=72, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c5n.xlarge': InstanceType(name='c5n.xlarge', cores=4, memory=10.5, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.12xlarge': InstanceType(name='c6a.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.16xlarge': InstanceType(name='c6a.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.24xlarge': InstanceType(name='c6a.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.2xlarge': InstanceType(name='c6a.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.32xlarge': InstanceType(name='c6a.32xlarge', cores=128, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.48xlarge': InstanceType(name='c6a.48xlarge', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.4xlarge': InstanceType(name='c6a.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.8xlarge': InstanceType(name='c6a.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.large': InstanceType(name='c6a.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.metal': InstanceType(name='c6a.metal', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6a.xlarge': InstanceType(name='c6a.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6g.12xlarge': InstanceType(name='c6g.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.16xlarge': InstanceType(name='c6g.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.2xlarge': InstanceType(name='c6g.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.4xlarge': InstanceType(name='c6g.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.8xlarge': InstanceType(name='c6g.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.large': InstanceType(name='c6g.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.medium': InstanceType(name='c6g.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.metal': InstanceType(name='c6g.metal', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6g.xlarge': InstanceType(name='c6g.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gd.12xlarge': InstanceType(name='c6gd.12xlarge', cores=48, memory=96.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'c6gd.16xlarge': InstanceType(name='c6gd.16xlarge', cores=64, memory=128.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'c6gd.2xlarge': InstanceType(name='c6gd.2xlarge', cores=8, memory=16.0, disks=1.0, disk_capacity=475.0, architecture='arm64'),
    'c6gd.4xlarge': InstanceType(name='c6gd.4xlarge', cores=16, memory=32.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'c6gd.8xlarge': InstanceType(name='c6gd.8xlarge', cores=32, memory=64.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'c6gd.large': InstanceType(name='c6gd.large', cores=2, memory=4.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'c6gd.medium': InstanceType(name='c6gd.medium', cores=1, memory=2.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'c6gd.metal': InstanceType(name='c6gd.metal', cores=64, memory=128.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'c6gd.xlarge': InstanceType(name='c6gd.xlarge', cores=4, memory=8.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'c6gn.12xlarge': InstanceType(name='c6gn.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.16xlarge': InstanceType(name='c6gn.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.2xlarge': InstanceType(name='c6gn.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.4xlarge': InstanceType(name='c6gn.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.8xlarge': InstanceType(name='c6gn.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.large': InstanceType(name='c6gn.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.medium': InstanceType(name='c6gn.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.metal': InstanceType(name='c6gn.metal', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6gn.xlarge': InstanceType(name='c6gn.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c6i.12xlarge': InstanceType(name='c6i.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.16xlarge': InstanceType(name='c6i.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.24xlarge': InstanceType(name='c6i.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.2xlarge': InstanceType(name='c6i.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.32xlarge': InstanceType(name='c6i.32xlarge', cores=128, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.4xlarge': InstanceType(name='c6i.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.8xlarge': InstanceType(name='c6i.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.large': InstanceType(name='c6i.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.metal': InstanceType(name='c6i.metal', cores=128, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6i.xlarge': InstanceType(name='c6i.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6id.12xlarge': InstanceType(name='c6id.12xlarge', cores=48, memory=96.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'c6id.16xlarge': InstanceType(name='c6id.16xlarge', cores=64, memory=128.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'c6id.24xlarge': InstanceType(name='c6id.24xlarge', cores=96, memory=192.0, disks=4.0, disk_capacity=1425.0, architecture='amd64'),
    'c6id.2xlarge': InstanceType(name='c6id.2xlarge', cores=8, memory=16.0, disks=1.0, disk_capacity=474.0, architecture='amd64'),
    'c6id.32xlarge': InstanceType(name='c6id.32xlarge', cores=128, memory=256.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'c6id.4xlarge': InstanceType(name='c6id.4xlarge', cores=16, memory=32.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'c6id.8xlarge': InstanceType(name='c6id.8xlarge', cores=32, memory=64.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'c6id.large': InstanceType(name='c6id.large', cores=2, memory=4.0, disks=1.0, disk_capacity=118.0, architecture='amd64'),
    'c6id.metal': InstanceType(name='c6id.metal', cores=128, memory=256.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'c6id.xlarge': InstanceType(name='c6id.xlarge', cores=4, memory=8.0, disks=1.0, disk_capacity=237.0, architecture='amd64'),
    'c6in.12xlarge': InstanceType(name='c6in.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.16xlarge': InstanceType(name='c6in.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.24xlarge': InstanceType(name='c6in.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.2xlarge': InstanceType(name='c6in.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.32xlarge': InstanceType(name='c6in.32xlarge', cores=128, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.4xlarge': InstanceType(name='c6in.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.8xlarge': InstanceType(name='c6in.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.large': InstanceType(name='c6in.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.metal': InstanceType(name='c6in.metal', cores=128, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c6in.xlarge': InstanceType(name='c6in.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.12xlarge': InstanceType(name='c7a.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.16xlarge': InstanceType(name='c7a.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.24xlarge': InstanceType(name='c7a.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.2xlarge': InstanceType(name='c7a.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.32xlarge': InstanceType(name='c7a.32xlarge', cores=128, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.48xlarge': InstanceType(name='c7a.48xlarge', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.4xlarge': InstanceType(name='c7a.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.8xlarge': InstanceType(name='c7a.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.large': InstanceType(name='c7a.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.medium': InstanceType(name='c7a.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.metal-48xl': InstanceType(name='c7a.metal-48xl', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7a.xlarge': InstanceType(name='c7a.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7g.12xlarge': InstanceType(name='c7g.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.16xlarge': InstanceType(name='c7g.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.2xlarge': InstanceType(name='c7g.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.4xlarge': InstanceType(name='c7g.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.8xlarge': InstanceType(name='c7g.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.large': InstanceType(name='c7g.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.medium': InstanceType(name='c7g.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.metal': InstanceType(name='c7g.metal', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7g.xlarge': InstanceType(name='c7g.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gd.12xlarge': InstanceType(name='c7gd.12xlarge', cores=48, memory=96.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'c7gd.16xlarge': InstanceType(name='c7gd.16xlarge', cores=64, memory=128.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'c7gd.2xlarge': InstanceType(name='c7gd.2xlarge', cores=8, memory=16.0, disks=1.0, disk_capacity=475.0, architecture='arm64'),
    'c7gd.4xlarge': InstanceType(name='c7gd.4xlarge', cores=16, memory=32.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'c7gd.8xlarge': InstanceType(name='c7gd.8xlarge', cores=32, memory=64.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'c7gd.large': InstanceType(name='c7gd.large', cores=2, memory=4.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'c7gd.medium': InstanceType(name='c7gd.medium', cores=1, memory=2.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'c7gd.metal': InstanceType(name='c7gd.metal', cores=64, memory=128.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'c7gd.xlarge': InstanceType(name='c7gd.xlarge', cores=4, memory=8.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'c7gn.12xlarge': InstanceType(name='c7gn.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.16xlarge': InstanceType(name='c7gn.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.2xlarge': InstanceType(name='c7gn.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.4xlarge': InstanceType(name='c7gn.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.8xlarge': InstanceType(name='c7gn.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.large': InstanceType(name='c7gn.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.medium': InstanceType(name='c7gn.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.metal': InstanceType(name='c7gn.metal', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7gn.xlarge': InstanceType(name='c7gn.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c7i-flex.12xlarge': InstanceType(name='c7i-flex.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i-flex.16xlarge': InstanceType(name='c7i-flex.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i-flex.2xlarge': InstanceType(name='c7i-flex.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i-flex.4xlarge': InstanceType(name='c7i-flex.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i-flex.8xlarge': InstanceType(name='c7i-flex.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i-flex.large': InstanceType(name='c7i-flex.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i-flex.xlarge': InstanceType(name='c7i-flex.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.12xlarge': InstanceType(name='c7i.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.16xlarge': InstanceType(name='c7i.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.24xlarge': InstanceType(name='c7i.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.2xlarge': InstanceType(name='c7i.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.48xlarge': InstanceType(name='c7i.48xlarge', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.4xlarge': InstanceType(name='c7i.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.8xlarge': InstanceType(name='c7i.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.large': InstanceType(name='c7i.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.metal-24xl': InstanceType(name='c7i.metal-24xl', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.metal-48xl': InstanceType(name='c7i.metal-48xl', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c7i.xlarge': InstanceType(name='c7i.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'c8g.12xlarge': InstanceType(name='c8g.12xlarge', cores=48, memory=96.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.16xlarge': InstanceType(name='c8g.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.24xlarge': InstanceType(name='c8g.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.2xlarge': InstanceType(name='c8g.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.48xlarge': InstanceType(name='c8g.48xlarge', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.4xlarge': InstanceType(name='c8g.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.8xlarge': InstanceType(name='c8g.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.large': InstanceType(name='c8g.large', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.medium': InstanceType(name='c8g.medium', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.metal-24xl': InstanceType(name='c8g.metal-24xl', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.metal-48xl': InstanceType(name='c8g.metal-48xl', cores=192, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'c8g.xlarge': InstanceType(name='c8g.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'cc2.8xlarge': InstanceType(name='cc2.8xlarge', cores=32, memory=60.5, disks=4.0, disk_capacity=840.0, architecture='amd64'),
    'cr1.8xlarge': InstanceType(name='cr1.8xlarge', cores=32, memory=244.0, disks=2.0, disk_capacity=120.0, architecture='amd64'),
    'd2.2xlarge': InstanceType(name='d2.2xlarge', cores=8, memory=61.0, disks=6.0, disk_capacity=2000.0, architecture='amd64'),
    'd2.4xlarge': InstanceType(name='d2.4xlarge', cores=16, memory=122.0, disks=12.0, disk_capacity=2000.0, architecture='amd64'),
    'd2.8xlarge': InstanceType(name='d2.8xlarge', cores=36, memory=244.0, disks=24.0, disk_capacity=2000.0, architecture='amd64'),
    'd2.xlarge': InstanceType(name='d2.xlarge', cores=4, memory=30.5, disks=3.0, disk_capacity=2000.0, architecture='amd64'),
    'd3.2xlarge': InstanceType(name='d3.2xlarge', cores=8, memory=64.0, disks=6.0, disk_capacity=2000.0, architecture='amd64'),
    'd3.4xlarge': InstanceType(name='d3.4xlarge', cores=16, memory=128.0, disks=12.0, disk_capacity=2000.0, architecture='amd64'),
    'd3.8xlarge': InstanceType(name='d3.8xlarge', cores=32, memory=256.0, disks=24.0, disk_capacity=2000.0, architecture='amd64'),
    'd3.xlarge': InstanceType(name='d3.xlarge', cores=4, memory=32.0, disks=3.0, disk_capacity=2000.0, architecture='amd64'),
    'd3en.12xlarge': InstanceType(name='d3en.12xlarge', cores=48, memory=192.0, disks=24.0, disk_capacity=14000.0, architecture='amd64'),
    'd3en.2xlarge': InstanceType(name='d3en.2xlarge', cores=8, memory=32.0, disks=4.0, disk_capacity=14000.0, architecture='amd64'),
    'd3en.4xlarge': InstanceType(name='d3en.4xlarge', cores=16, memory=64.0, disks=8.0, disk_capacity=14000.0, architecture='amd64'),
    'd3en.6xlarge': InstanceType(name='d3en.6xlarge', cores=24, memory=96.0, disks=12.0, disk_capacity=14000.0, architecture='amd64'),
    'd3en.8xlarge': InstanceType(name='d3en.8xlarge', cores=32, memory=128.0, disks=16.0, disk_capacity=14000.0, architecture='amd64'),
    'd3en.xlarge': InstanceType(name='d3en.xlarge', cores=4, memory=16.0, disks=2.0, disk_capacity=14000.0, architecture='amd64'),
    'dl1.24xlarge': InstanceType(name='dl1.24xlarge', cores=96, memory=768.0, disks=4.0, disk_capacity=1000.0, architecture='amd64'),
    'dl2q.24xlarge': InstanceType(name='dl2q.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'f1.16xlarge': InstanceType(name='f1.16xlarge', cores=64, memory=976.0, disks=4.0, disk_capacity=940.0, architecture='amd64'),
    'f1.2xlarge': InstanceType(name='f1.2xlarge', cores=8, memory=122.0, disks=1.0, disk_capacity=470.0, architecture='amd64'),
    'f1.4xlarge': InstanceType(name='f1.4xlarge', cores=16, memory=244.0, disks=1.0, disk_capacity=940.0, architecture='amd64'),
    'f2.12xlarge': InstanceType(name='f2.12xlarge', cores=48, memory=512.0, disks=2.0, disk_capacity=950.0, architecture='amd64'),
    'f2.48xlarge': InstanceType(name='f2.48xlarge', cores=192, memory=2048.0, disks=8.0, disk_capacity=950.0, architecture='amd64'),
    'f2.6xlarge': InstanceType(name='f2.6xlarge', cores=24, memory=256.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'g2.2xlarge': InstanceType(name='g2.2xlarge', cores=8, memory=15.0, disks=1.0, disk_capacity=60.0, architecture='amd64'),
    'g2.8xlarge': InstanceType(name='g2.8xlarge', cores=32, memory=60.0, disks=2.0, disk_capacity=120.0, architecture='amd64'),
    'g3.16xlarge': InstanceType(name='g3.16xlarge', cores=64, memory=488.0, disks=0, disk_capacity=0, architecture='amd64'),
    'g3.4xlarge': InstanceType(name='g3.4xlarge', cores=16, memory=122.0, disks=0, disk_capacity=0, architecture='amd64'),
    'g3.8xlarge': InstanceType(name='g3.8xlarge', cores=32, memory=244.0, disks=0, disk_capacity=0, architecture='amd64'),
    'g3s.xlarge': InstanceType(name='g3s.xlarge', cores=4, memory=30.5, disks=0, disk_capacity=0, architecture='amd64'),
    'g4ad.16xlarge': InstanceType(name='g4ad.16xlarge', cores=64, memory=256.0, disks=1, disk_capacity=2400.0, architecture='amd64'),
    'g4ad.2xlarge': InstanceType(name='g4ad.2xlarge', cores=8, memory=32.0, disks=1, disk_capacity=300.0, architecture='amd64'),
    'g4ad.4xlarge': InstanceType(name='g4ad.4xlarge', cores=16, memory=64.0, disks=1, disk_capacity=600.0, architecture='amd64'),
    'g4ad.8xlarge': InstanceType(name='g4ad.8xlarge', cores=32, memory=128.0, disks=1, disk_capacity=1200.0, architecture='amd64'),
    'g4ad.xlarge': InstanceType(name='g4ad.xlarge', cores=4, memory=16.0, disks=1, disk_capacity=150.0, architecture='amd64'),
    'g4dn.12xlarge': InstanceType(name='g4dn.12xlarge', cores=48, memory=192.0, disks=1, disk_capacity=900.0, architecture='amd64'),
    'g4dn.16xlarge': InstanceType(name='g4dn.16xlarge', cores=64, memory=256.0, disks=1, disk_capacity=900.0, architecture='amd64'),
    'g4dn.2xlarge': InstanceType(name='g4dn.2xlarge', cores=8, memory=32.0, disks=1, disk_capacity=225.0, architecture='amd64'),
    'g4dn.4xlarge': InstanceType(name='g4dn.4xlarge', cores=16, memory=64.0, disks=1, disk_capacity=225.0, architecture='amd64'),
    'g4dn.8xlarge': InstanceType(name='g4dn.8xlarge', cores=32, memory=128.0, disks=1, disk_capacity=900.0, architecture='amd64'),
    'g4dn.metal': InstanceType(name='g4dn.metal', cores=96, memory=384.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'g4dn.xlarge': InstanceType(name='g4dn.xlarge', cores=4, memory=16.0, disks=1, disk_capacity=125.0, architecture='amd64'),
    'g5.12xlarge': InstanceType(name='g5.12xlarge', cores=48, memory=192.0, disks=1.0, disk_capacity=3800.0, architecture='amd64'),
    'g5.16xlarge': InstanceType(name='g5.16xlarge', cores=64, memory=256.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'g5.24xlarge': InstanceType(name='g5.24xlarge', cores=96, memory=384.0, disks=1.0, disk_capacity=3800.0, architecture='amd64'),
    'g5.2xlarge': InstanceType(name='g5.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=450.0, architecture='amd64'),
    'g5.48xlarge': InstanceType(name='g5.48xlarge', cores=192, memory=768.0, disks=2.0, disk_capacity=3800.0, architecture='amd64'),
    'g5.4xlarge': InstanceType(name='g5.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=600.0, architecture='amd64'),
    'g5.8xlarge': InstanceType(name='g5.8xlarge', cores=32, memory=128.0, disks=1.0, disk_capacity=900.0, architecture='amd64'),
    'g5.xlarge': InstanceType(name='g5.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=250.0, architecture='amd64'),
    'g5g.16xlarge': InstanceType(name='g5g.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'g5g.2xlarge': InstanceType(name='g5g.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'g5g.4xlarge': InstanceType(name='g5g.4xlarge', cores=16, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'g5g.8xlarge': InstanceType(name='g5g.8xlarge', cores=32, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'g5g.metal': InstanceType(name='g5g.metal', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'g5g.xlarge': InstanceType(name='g5g.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'g6.12xlarge': InstanceType(name='g6.12xlarge', cores=48, memory=192.0, disks=4.0, disk_capacity=940.0, architecture='amd64'),
    'g6.16xlarge': InstanceType(name='g6.16xlarge', cores=64, memory=256.0, disks=2.0, disk_capacity=940.0, architecture='amd64'),
    'g6.24xlarge': InstanceType(name='g6.24xlarge', cores=96, memory=384.0, disks=4.0, disk_capacity=940.0, architecture='amd64'),
    'g6.2xlarge': InstanceType(name='g6.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=450.0, architecture='amd64'),
    'g6.48xlarge': InstanceType(name='g6.48xlarge', cores=192, memory=768.0, disks=8.0, disk_capacity=940.0, architecture='amd64'),
    'g6.4xlarge': InstanceType(name='g6.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=600.0, architecture='amd64'),
    'g6.8xlarge': InstanceType(name='g6.8xlarge', cores=32, memory=128.0, disks=2.0, disk_capacity=450.0, architecture='amd64'),
    'g6.xlarge': InstanceType(name='g6.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=250.0, architecture='amd64'),
    'g6e.12xlarge': InstanceType(name='g6e.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'g6e.16xlarge': InstanceType(name='g6e.16xlarge', cores=64, memory=512.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'g6e.24xlarge': InstanceType(name='g6e.24xlarge', cores=96, memory=768.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'g6e.2xlarge': InstanceType(name='g6e.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=450.0, architecture='amd64'),
    'g6e.48xlarge': InstanceType(name='g6e.48xlarge', cores=192, memory=1536.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'g6e.4xlarge': InstanceType(name='g6e.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=600.0, architecture='amd64'),
    'g6e.8xlarge': InstanceType(name='g6e.8xlarge', cores=32, memory=256.0, disks=1.0, disk_capacity=900.0, architecture='amd64'),
    'g6e.xlarge': InstanceType(name='g6e.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=250.0, architecture='amd64'),
    'gr6.4xlarge': InstanceType(name='gr6.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=600.0, architecture='amd64'),
    'gr6.8xlarge': InstanceType(name='gr6.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=450.0, architecture='amd64'),
    'h1.16xlarge': InstanceType(name='h1.16xlarge', cores=64, memory=256.0, disks=8.0, disk_capacity=2000.0, architecture='amd64'),
    'h1.2xlarge': InstanceType(name='h1.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=2000.0, architecture='amd64'),
    'h1.4xlarge': InstanceType(name='h1.4xlarge', cores=16, memory=64.0, disks=2.0, disk_capacity=2000.0, architecture='amd64'),
    'h1.8xlarge': InstanceType(name='h1.8xlarge', cores=32, memory=128.0, disks=4.0, disk_capacity=2000.0, architecture='amd64'),
    'hpc6a.48xlarge': InstanceType(name='hpc6a.48xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'hpc6id.32xlarge': InstanceType(name='hpc6id.32xlarge', cores=64, memory=1024.0, disks=4.0, disk_capacity=3800.0, architecture='amd64'),
    'hpc7a.12xlarge': InstanceType(name='hpc7a.12xlarge', cores=24, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'hpc7a.24xlarge': InstanceType(name='hpc7a.24xlarge', cores=48, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'hpc7a.48xlarge': InstanceType(name='hpc7a.48xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'hpc7a.96xlarge': InstanceType(name='hpc7a.96xlarge', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'hpc7g.16xlarge': InstanceType(name='hpc7g.16xlarge', cores=64, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'hpc7g.4xlarge': InstanceType(name='hpc7g.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'hpc7g.8xlarge': InstanceType(name='hpc7g.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'hs1.8xlarge': InstanceType(name='hs1.8xlarge', cores=16, memory=117.0, disks=24.0, disk_capacity=2000.0, architecture='amd64'),
    'i2.2xlarge': InstanceType(name='i2.2xlarge', cores=8, memory=61.0, disks=2.0, disk_capacity=800.0, architecture='amd64'),
    'i2.4xlarge': InstanceType(name='i2.4xlarge', cores=16, memory=122.0, disks=4.0, disk_capacity=800.0, architecture='amd64'),
    'i2.8xlarge': InstanceType(name='i2.8xlarge', cores=32, memory=244.0, disks=8.0, disk_capacity=800.0, architecture='amd64'),
    'i2.xlarge': InstanceType(name='i2.xlarge', cores=4, memory=30.5, disks=1.0, disk_capacity=800.0, architecture='amd64'),
    'i3.16xlarge': InstanceType(name='i3.16xlarge', cores=64, memory=488.0, disks=8.0, disk_capacity=1900.0, architecture='amd64'),
    'i3.2xlarge': InstanceType(name='i3.2xlarge', cores=8, memory=61.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'i3.4xlarge': InstanceType(name='i3.4xlarge', cores=16, memory=122.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'i3.8xlarge': InstanceType(name='i3.8xlarge', cores=32, memory=244.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'i3.large': InstanceType(name='i3.large', cores=2, memory=15.25, disks=1.0, disk_capacity=475.0, architecture='amd64'),
    'i3.metal': InstanceType(name='i3.metal', cores=72, memory=512.0, disks=8.0, disk_capacity=1900.0, architecture='amd64'),
    'i3.xlarge': InstanceType(name='i3.xlarge', cores=4, memory=30.5, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'i3en.12xlarge': InstanceType(name='i3en.12xlarge', cores=48, memory=384.0, disks=4.0, disk_capacity=7500.0, architecture='amd64'),
    'i3en.24xlarge': InstanceType(name='i3en.24xlarge', cores=96, memory=768.0, disks=8.0, disk_capacity=7500.0, architecture='amd64'),
    'i3en.2xlarge': InstanceType(name='i3en.2xlarge', cores=8, memory=64.0, disks=2.0, disk_capacity=2500.0, architecture='amd64'),
    'i3en.3xlarge': InstanceType(name='i3en.3xlarge', cores=12, memory=96.0, disks=1.0, disk_capacity=7500.0, architecture='amd64'),
    'i3en.6xlarge': InstanceType(name='i3en.6xlarge', cores=24, memory=192.0, disks=2.0, disk_capacity=7500.0, architecture='amd64'),
    'i3en.large': InstanceType(name='i3en.large', cores=2, memory=16.0, disks=1.0, disk_capacity=1250.0, architecture='amd64'),
    'i3en.metal': InstanceType(name='i3en.metal', cores=96, memory=768.0, disks=8.0, disk_capacity=7500.0, architecture='amd64'),
    'i3en.xlarge': InstanceType(name='i3en.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=2500.0, architecture='amd64'),
    'i3p.16xlarge': InstanceType(name='i3p.16xlarge', cores=64, memory=488.0, disks=0, disk_capacity=0, architecture='amd64'),
    'i4g.16xlarge': InstanceType(name='i4g.16xlarge', cores=64, memory=512.0, disks=4.0, disk_capacity=3750.0, architecture='arm64'),
    'i4g.2xlarge': InstanceType(name='i4g.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=1875.0, architecture='arm64'),
    'i4g.4xlarge': InstanceType(name='i4g.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=3750.0, architecture='arm64'),
    'i4g.8xlarge': InstanceType(name='i4g.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=3750.0, architecture='arm64'),
    'i4g.large': InstanceType(name='i4g.large', cores=2, memory=16.0, disks=1.0, disk_capacity=468.0, architecture='arm64'),
    'i4g.xlarge': InstanceType(name='i4g.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=937.0, architecture='arm64'),
    'i4i.12xlarge': InstanceType(name='i4i.12xlarge', cores=48, memory=384.0, disks=3.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.16xlarge': InstanceType(name='i4i.16xlarge', cores=64, memory=512.0, disks=4.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.24xlarge': InstanceType(name='i4i.24xlarge', cores=96, memory=768.0, disks=6.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.2xlarge': InstanceType(name='i4i.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=1875.0, architecture='amd64'),
    'i4i.32xlarge': InstanceType(name='i4i.32xlarge', cores=128, memory=1024.0, disks=8.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.4xlarge': InstanceType(name='i4i.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.8xlarge': InstanceType(name='i4i.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.large': InstanceType(name='i4i.large', cores=2, memory=16.0, disks=1.0, disk_capacity=468.0, architecture='amd64'),
    'i4i.metal': InstanceType(name='i4i.metal', cores=128, memory=1024.0, disks=8.0, disk_capacity=3750.0, architecture='amd64'),
    'i4i.xlarge': InstanceType(name='i4i.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=937.0, architecture='amd64'),
    'i7ie.12xlarge': InstanceType(name='i7ie.12xlarge', cores=48, memory=384.0, disks=4.0, disk_capacity=7500.0, architecture='amd64'),
    'i7ie.18xlarge': InstanceType(name='i7ie.18xlarge', cores=72, memory=576.0, disks=6.0, disk_capacity=7500.0, architecture='amd64'),
    'i7ie.24xlarge': InstanceType(name='i7ie.24xlarge', cores=96, memory=768.0, disks=8.0, disk_capacity=7500.0, architecture='amd64'),
    'i7ie.2xlarge': InstanceType(name='i7ie.2xlarge', cores=8, memory=64.0, disks=2.0, disk_capacity=2500.0, architecture='amd64'),
    'i7ie.3xlarge': InstanceType(name='i7ie.3xlarge', cores=12, memory=96.0, disks=1.0, disk_capacity=7500.0, architecture='amd64'),
    'i7ie.48xlarge': InstanceType(name='i7ie.48xlarge', cores=192, memory=1536.0, disks=16.0, disk_capacity=7500.0, architecture='amd64'),
    'i7ie.6xlarge': InstanceType(name='i7ie.6xlarge', cores=24, memory=192.0, disks=2.0, disk_capacity=7500.0, architecture='amd64'),
    'i7ie.large': InstanceType(name='i7ie.large', cores=2, memory=16.0, disks=1.0, disk_capacity=1250.0, architecture='amd64'),
    'i7ie.xlarge': InstanceType(name='i7ie.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=2500.0, architecture='amd64'),
    'i8g.12xlarge': InstanceType(name='i8g.12xlarge', cores=48, memory=384.0, disks=3.0, disk_capacity=3750.0, architecture='arm64'),
    'i8g.16xlarge': InstanceType(name='i8g.16xlarge', cores=64, memory=512.0, disks=4.0, disk_capacity=3750.0, architecture='arm64'),
    'i8g.24xlarge': InstanceType(name='i8g.24xlarge', cores=96, memory=768.0, disks=6.0, disk_capacity=3750.0, architecture='arm64'),
    'i8g.2xlarge': InstanceType(name='i8g.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=1875.0, architecture='arm64'),
    'i8g.4xlarge': InstanceType(name='i8g.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=3750.0, architecture='arm64'),
    'i8g.8xlarge': InstanceType(name='i8g.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=3750.0, architecture='arm64'),
    'i8g.large': InstanceType(name='i8g.large', cores=2, memory=16.0, disks=1.0, disk_capacity=468.0, architecture='arm64'),
    'i8g.metal-24xl': InstanceType(name='i8g.metal-24xl', cores=96, memory=768.0, disks=6.0, disk_capacity=3750.0, architecture='arm64'),
    'i8g.xlarge': InstanceType(name='i8g.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=937.0, architecture='arm64'),
    'im4gn.16xlarge': InstanceType(name='im4gn.16xlarge', cores=64, memory=256.0, disks=4.0, disk_capacity=7500.0, architecture='arm64'),
    'im4gn.2xlarge': InstanceType(name='im4gn.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=3750.0, architecture='arm64'),
    'im4gn.4xlarge': InstanceType(name='im4gn.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=7500.0, architecture='arm64'),
    'im4gn.8xlarge': InstanceType(name='im4gn.8xlarge', cores=32, memory=128.0, disks=2.0, disk_capacity=7500.0, architecture='arm64'),
    'im4gn.large': InstanceType(name='im4gn.large', cores=2, memory=8.0, disks=1.0, disk_capacity=937.0, architecture='arm64'),
    'im4gn.xlarge': InstanceType(name='im4gn.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=1875.0, architecture='arm64'),
    'inf1.24xlarge': InstanceType(name='inf1.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf1.2xlarge': InstanceType(name='inf1.2xlarge', cores=8, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf1.6xlarge': InstanceType(name='inf1.6xlarge', cores=24, memory=48.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf1.xlarge': InstanceType(name='inf1.xlarge', cores=4, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf2.24xlarge': InstanceType(name='inf2.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf2.48xlarge': InstanceType(name='inf2.48xlarge', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf2.8xlarge': InstanceType(name='inf2.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'inf2.xlarge': InstanceType(name='inf2.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'is4gen.2xlarge': InstanceType(name='is4gen.2xlarge', cores=8, memory=48.0, disks=1.0, disk_capacity=7500.0, architecture='arm64'),
    'is4gen.4xlarge': InstanceType(name='is4gen.4xlarge', cores=16, memory=96.0, disks=2.0, disk_capacity=7500.0, architecture='arm64'),
    'is4gen.8xlarge': InstanceType(name='is4gen.8xlarge', cores=32, memory=192.0, disks=4.0, disk_capacity=7500.0, architecture='arm64'),
    'is4gen.large': InstanceType(name='is4gen.large', cores=2, memory=12.0, disks=1.0, disk_capacity=1875.0, architecture='arm64'),
    'is4gen.medium': InstanceType(name='is4gen.medium', cores=1, memory=6.0, disks=1.0, disk_capacity=937.0, architecture='arm64'),
    'is4gen.xlarge': InstanceType(name='is4gen.xlarge', cores=4, memory=24.0, disks=1.0, disk_capacity=3750.0, architecture='arm64'),
    'm1.large': InstanceType(name='m1.large', cores=2, memory=7.5, disks=2.0, disk_capacity=420.0, architecture='amd64'),
    'm1.medium': InstanceType(name='m1.medium', cores=1, memory=3.75, disks=1.0, disk_capacity=410.0, architecture='amd64'),
    'm1.xlarge': InstanceType(name='m1.xlarge', cores=4, memory=15.0, disks=4.0, disk_capacity=420.0, architecture='amd64'),
    'm2.2xlarge': InstanceType(name='m2.2xlarge', cores=4, memory=34.2, disks=1.0, disk_capacity=850.0, architecture='amd64'),
    'm2.4xlarge': InstanceType(name='m2.4xlarge', cores=8, memory=68.4, disks=2.0, disk_capacity=840.0, architecture='amd64'),
    'm2.xlarge': InstanceType(name='m2.xlarge', cores=2, memory=17.1, disks=1.0, disk_capacity=420.0, architecture='amd64'),
    'm3.2xlarge': InstanceType(name='m3.2xlarge', cores=8, memory=30.0, disks=2.0, disk_capacity=80.0, architecture='amd64'),
    'm3.large': InstanceType(name='m3.large', cores=2, memory=7.5, disks=1.0, disk_capacity=32.0, architecture='amd64'),
    'm3.medium': InstanceType(name='m3.medium', cores=1, memory=3.75, disks=1.0, disk_capacity=4.0, architecture='amd64'),
    'm3.xlarge': InstanceType(name='m3.xlarge', cores=4, memory=15.0, disks=2.0, disk_capacity=40.0, architecture='amd64'),
    'm4.10xlarge': InstanceType(name='m4.10xlarge', cores=40, memory=160.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm4.16xlarge': InstanceType(name='m4.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm4.2xlarge': InstanceType(name='m4.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm4.4xlarge': InstanceType(name='m4.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm4.large': InstanceType(name='m4.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm4.xlarge': InstanceType(name='m4.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.12xlarge': InstanceType(name='m5.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.16xlarge': InstanceType(name='m5.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.24xlarge': InstanceType(name='m5.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.2xlarge': InstanceType(name='m5.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.4xlarge': InstanceType(name='m5.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.8xlarge': InstanceType(name='m5.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.large': InstanceType(name='m5.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.metal': InstanceType(name='m5.metal', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5.xlarge': InstanceType(name='m5.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.12xlarge': InstanceType(name='m5a.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.16xlarge': InstanceType(name='m5a.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.24xlarge': InstanceType(name='m5a.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.2xlarge': InstanceType(name='m5a.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.4xlarge': InstanceType(name='m5a.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.8xlarge': InstanceType(name='m5a.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.large': InstanceType(name='m5a.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5a.xlarge': InstanceType(name='m5a.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5ad.12xlarge': InstanceType(name='m5ad.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'm5ad.16xlarge': InstanceType(name='m5ad.16xlarge', cores=64, memory=256.0, disks=4.0, disk_capacity=600.0, architecture='amd64'),
    'm5ad.24xlarge': InstanceType(name='m5ad.24xlarge', cores=96, memory=384.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'm5ad.2xlarge': InstanceType(name='m5ad.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'm5ad.4xlarge': InstanceType(name='m5ad.4xlarge', cores=16, memory=64.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'm5ad.8xlarge': InstanceType(name='m5ad.8xlarge', cores=32, memory=128.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'm5ad.large': InstanceType(name='m5ad.large', cores=2, memory=8.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'm5ad.xlarge': InstanceType(name='m5ad.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'm5d.12xlarge': InstanceType(name='m5d.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'm5d.16xlarge': InstanceType(name='m5d.16xlarge', cores=64, memory=256.0, disks=4.0, disk_capacity=600.0, architecture='amd64'),
    'm5d.24xlarge': InstanceType(name='m5d.24xlarge', cores=96, memory=384.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'm5d.2xlarge': InstanceType(name='m5d.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'm5d.4xlarge': InstanceType(name='m5d.4xlarge', cores=16, memory=64.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'm5d.8xlarge': InstanceType(name='m5d.8xlarge', cores=32, memory=128.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'm5d.large': InstanceType(name='m5d.large', cores=2, memory=8.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'm5d.metal': InstanceType(name='m5d.metal', cores=96, memory=384.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'm5d.xlarge': InstanceType(name='m5d.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'm5dn.12xlarge': InstanceType(name='m5dn.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'm5dn.16xlarge': InstanceType(name='m5dn.16xlarge', cores=64, memory=256.0, disks=4.0, disk_capacity=600.0, architecture='amd64'),
    'm5dn.24xlarge': InstanceType(name='m5dn.24xlarge', cores=96, memory=384.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'm5dn.2xlarge': InstanceType(name='m5dn.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'm5dn.4xlarge': InstanceType(name='m5dn.4xlarge', cores=16, memory=64.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'm5dn.8xlarge': InstanceType(name='m5dn.8xlarge', cores=32, memory=128.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'm5dn.large': InstanceType(name='m5dn.large', cores=2, memory=8.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'm5dn.metal': InstanceType(name='m5dn.metal', cores=96, memory=384.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'm5dn.xlarge': InstanceType(name='m5dn.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'm5n.12xlarge': InstanceType(name='m5n.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.16xlarge': InstanceType(name='m5n.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.24xlarge': InstanceType(name='m5n.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.2xlarge': InstanceType(name='m5n.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.4xlarge': InstanceType(name='m5n.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.8xlarge': InstanceType(name='m5n.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.large': InstanceType(name='m5n.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.metal': InstanceType(name='m5n.metal', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5n.xlarge': InstanceType(name='m5n.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.12xlarge': InstanceType(name='m5zn.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.2xlarge': InstanceType(name='m5zn.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.3xlarge': InstanceType(name='m5zn.3xlarge', cores=12, memory=48.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.6xlarge': InstanceType(name='m5zn.6xlarge', cores=24, memory=96.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.large': InstanceType(name='m5zn.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.metal': InstanceType(name='m5zn.metal', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm5zn.xlarge': InstanceType(name='m5zn.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.12xlarge': InstanceType(name='m6a.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.16xlarge': InstanceType(name='m6a.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.24xlarge': InstanceType(name='m6a.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.2xlarge': InstanceType(name='m6a.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.32xlarge': InstanceType(name='m6a.32xlarge', cores=128, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.48xlarge': InstanceType(name='m6a.48xlarge', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.4xlarge': InstanceType(name='m6a.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.8xlarge': InstanceType(name='m6a.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.large': InstanceType(name='m6a.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.metal': InstanceType(name='m6a.metal', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6a.xlarge': InstanceType(name='m6a.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6g.12xlarge': InstanceType(name='m6g.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.16xlarge': InstanceType(name='m6g.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.2xlarge': InstanceType(name='m6g.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.4xlarge': InstanceType(name='m6g.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.8xlarge': InstanceType(name='m6g.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.large': InstanceType(name='m6g.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.medium': InstanceType(name='m6g.medium', cores=1, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.metal': InstanceType(name='m6g.metal', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6g.xlarge': InstanceType(name='m6g.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm6gd.12xlarge': InstanceType(name='m6gd.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'm6gd.16xlarge': InstanceType(name='m6gd.16xlarge', cores=64, memory=256.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'm6gd.2xlarge': InstanceType(name='m6gd.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=475.0, architecture='arm64'),
    'm6gd.4xlarge': InstanceType(name='m6gd.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'm6gd.8xlarge': InstanceType(name='m6gd.8xlarge', cores=32, memory=128.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'm6gd.large': InstanceType(name='m6gd.large', cores=2, memory=8.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'm6gd.medium': InstanceType(name='m6gd.medium', cores=1, memory=4.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'm6gd.metal': InstanceType(name='m6gd.metal', cores=64, memory=256.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'm6gd.xlarge': InstanceType(name='m6gd.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'm6i.12xlarge': InstanceType(name='m6i.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.16xlarge': InstanceType(name='m6i.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.24xlarge': InstanceType(name='m6i.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.2xlarge': InstanceType(name='m6i.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.32xlarge': InstanceType(name='m6i.32xlarge', cores=128, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.4xlarge': InstanceType(name='m6i.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.8xlarge': InstanceType(name='m6i.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.large': InstanceType(name='m6i.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.metal': InstanceType(name='m6i.metal', cores=128, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6i.xlarge': InstanceType(name='m6i.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6id.12xlarge': InstanceType(name='m6id.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'm6id.16xlarge': InstanceType(name='m6id.16xlarge', cores=64, memory=256.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'm6id.24xlarge': InstanceType(name='m6id.24xlarge', cores=96, memory=384.0, disks=4.0, disk_capacity=1425.0, architecture='amd64'),
    'm6id.2xlarge': InstanceType(name='m6id.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=474.0, architecture='amd64'),
    'm6id.32xlarge': InstanceType(name='m6id.32xlarge', cores=128, memory=512.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'm6id.4xlarge': InstanceType(name='m6id.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'm6id.8xlarge': InstanceType(name='m6id.8xlarge', cores=32, memory=128.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'm6id.large': InstanceType(name='m6id.large', cores=2, memory=8.0, disks=1.0, disk_capacity=118.0, architecture='amd64'),
    'm6id.metal': InstanceType(name='m6id.metal', cores=128, memory=512.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'm6id.xlarge': InstanceType(name='m6id.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=237.0, architecture='amd64'),
    'm6idn.12xlarge': InstanceType(name='m6idn.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'm6idn.16xlarge': InstanceType(name='m6idn.16xlarge', cores=64, memory=256.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'm6idn.24xlarge': InstanceType(name='m6idn.24xlarge', cores=96, memory=384.0, disks=4.0, disk_capacity=1425.0, architecture='amd64'),
    'm6idn.2xlarge': InstanceType(name='m6idn.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=474.0, architecture='amd64'),
    'm6idn.32xlarge': InstanceType(name='m6idn.32xlarge', cores=128, memory=512.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'm6idn.4xlarge': InstanceType(name='m6idn.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'm6idn.8xlarge': InstanceType(name='m6idn.8xlarge', cores=32, memory=128.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'm6idn.large': InstanceType(name='m6idn.large', cores=2, memory=8.0, disks=1.0, disk_capacity=118.0, architecture='amd64'),
    'm6idn.metal': InstanceType(name='m6idn.metal', cores=128, memory=512.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'm6idn.xlarge': InstanceType(name='m6idn.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=237.0, architecture='amd64'),
    'm6in.12xlarge': InstanceType(name='m6in.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.16xlarge': InstanceType(name='m6in.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.24xlarge': InstanceType(name='m6in.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.2xlarge': InstanceType(name='m6in.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.32xlarge': InstanceType(name='m6in.32xlarge', cores=128, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.4xlarge': InstanceType(name='m6in.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.8xlarge': InstanceType(name='m6in.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.large': InstanceType(name='m6in.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.metal': InstanceType(name='m6in.metal', cores=128, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm6in.xlarge': InstanceType(name='m6in.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.12xlarge': InstanceType(name='m7a.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.16xlarge': InstanceType(name='m7a.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.24xlarge': InstanceType(name='m7a.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.2xlarge': InstanceType(name='m7a.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.32xlarge': InstanceType(name='m7a.32xlarge', cores=128, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.48xlarge': InstanceType(name='m7a.48xlarge', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.4xlarge': InstanceType(name='m7a.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.8xlarge': InstanceType(name='m7a.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.large': InstanceType(name='m7a.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.medium': InstanceType(name='m7a.medium', cores=1, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.metal-48xl': InstanceType(name='m7a.metal-48xl', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7a.xlarge': InstanceType(name='m7a.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7g.12xlarge': InstanceType(name='m7g.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.16xlarge': InstanceType(name='m7g.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.2xlarge': InstanceType(name='m7g.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.4xlarge': InstanceType(name='m7g.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.8xlarge': InstanceType(name='m7g.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.large': InstanceType(name='m7g.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.medium': InstanceType(name='m7g.medium', cores=1, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.metal': InstanceType(name='m7g.metal', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7g.xlarge': InstanceType(name='m7g.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm7gd.12xlarge': InstanceType(name='m7gd.12xlarge', cores=48, memory=192.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'm7gd.16xlarge': InstanceType(name='m7gd.16xlarge', cores=64, memory=256.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'm7gd.2xlarge': InstanceType(name='m7gd.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=475.0, architecture='arm64'),
    'm7gd.4xlarge': InstanceType(name='m7gd.4xlarge', cores=16, memory=64.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'm7gd.8xlarge': InstanceType(name='m7gd.8xlarge', cores=32, memory=128.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'm7gd.large': InstanceType(name='m7gd.large', cores=2, memory=8.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'm7gd.medium': InstanceType(name='m7gd.medium', cores=1, memory=4.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'm7gd.metal': InstanceType(name='m7gd.metal', cores=64, memory=256.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'm7gd.xlarge': InstanceType(name='m7gd.xlarge', cores=4, memory=16.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'm7i-flex.12xlarge': InstanceType(name='m7i-flex.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i-flex.16xlarge': InstanceType(name='m7i-flex.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i-flex.2xlarge': InstanceType(name='m7i-flex.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i-flex.4xlarge': InstanceType(name='m7i-flex.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i-flex.8xlarge': InstanceType(name='m7i-flex.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i-flex.large': InstanceType(name='m7i-flex.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i-flex.xlarge': InstanceType(name='m7i-flex.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.12xlarge': InstanceType(name='m7i.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.16xlarge': InstanceType(name='m7i.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.24xlarge': InstanceType(name='m7i.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.2xlarge': InstanceType(name='m7i.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.48xlarge': InstanceType(name='m7i.48xlarge', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.4xlarge': InstanceType(name='m7i.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.8xlarge': InstanceType(name='m7i.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.large': InstanceType(name='m7i.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.metal-24xl': InstanceType(name='m7i.metal-24xl', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.metal-48xl': InstanceType(name='m7i.metal-48xl', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm7i.xlarge': InstanceType(name='m7i.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'm8g.12xlarge': InstanceType(name='m8g.12xlarge', cores=48, memory=192.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.16xlarge': InstanceType(name='m8g.16xlarge', cores=64, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.24xlarge': InstanceType(name='m8g.24xlarge', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.2xlarge': InstanceType(name='m8g.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.48xlarge': InstanceType(name='m8g.48xlarge', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.4xlarge': InstanceType(name='m8g.4xlarge', cores=16, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.8xlarge': InstanceType(name='m8g.8xlarge', cores=32, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.large': InstanceType(name='m8g.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.medium': InstanceType(name='m8g.medium', cores=1, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.metal-24xl': InstanceType(name='m8g.metal-24xl', cores=96, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.metal-48xl': InstanceType(name='m8g.metal-48xl', cores=192, memory=768.0, disks=0, disk_capacity=0, architecture='arm64'),
    'm8g.xlarge': InstanceType(name='m8g.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'p2.16xlarge': InstanceType(name='p2.16xlarge', cores=64, memory=732.0, disks=0, disk_capacity=0, architecture='amd64'),
    'p2.8xlarge': InstanceType(name='p2.8xlarge', cores=32, memory=488.0, disks=0, disk_capacity=0, architecture='amd64'),
    'p2.xlarge': InstanceType(name='p2.xlarge', cores=4, memory=61.0, disks=0, disk_capacity=0, architecture='amd64'),
    'p3.16xlarge': InstanceType(name='p3.16xlarge', cores=64, memory=488.0, disks=0, disk_capacity=0, architecture='amd64'),
    'p3.2xlarge': InstanceType(name='p3.2xlarge', cores=8, memory=61.0, disks=0, disk_capacity=0, architecture='amd64'),
    'p3.8xlarge': InstanceType(name='p3.8xlarge', cores=32, memory=244.0, disks=0, disk_capacity=0, architecture='amd64'),
    'p3dn.24xlarge': InstanceType(name='p3dn.24xlarge', cores=96, memory=768.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'p4d.24xlarge': InstanceType(name='p4d.24xlarge', cores=96, memory=1152.0, disks=8.0, disk_capacity=1000.0, architecture='amd64'),
    'p4de.24xlarge': InstanceType(name='p4de.24xlarge', cores=96, memory=1152.0, disks=8.0, disk_capacity=1000.0, architecture='amd64'),
    'p5.48xlarge': InstanceType(name='p5.48xlarge', cores=192, memory=2048.0, disks=8.0, disk_capacity=3840.0, architecture='amd64'),
    'p5en.48xlarge': InstanceType(name='p5en.48xlarge', cores=192, memory=2048.0, disks=8.0, disk_capacity=3840.0, architecture='amd64'),
    'r3.2xlarge': InstanceType(name='r3.2xlarge', cores=8, memory=61.0, disks=1.0, disk_capacity=160.0, architecture='amd64'),
    'r3.4xlarge': InstanceType(name='r3.4xlarge', cores=16, memory=122.0, disks=1.0, disk_capacity=320.0, architecture='amd64'),
    'r3.8xlarge': InstanceType(name='r3.8xlarge', cores=32, memory=244.0, disks=2.0, disk_capacity=320.0, architecture='amd64'),
    'r3.large': InstanceType(name='r3.large', cores=2, memory=15.25, disks=1.0, disk_capacity=32.0, architecture='amd64'),
    'r3.xlarge': InstanceType(name='r3.xlarge', cores=4, memory=30.5, disks=1.0, disk_capacity=80.0, architecture='amd64'),
    'r4.16xlarge': InstanceType(name='r4.16xlarge', cores=64, memory=488.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r4.2xlarge': InstanceType(name='r4.2xlarge', cores=8, memory=61.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r4.4xlarge': InstanceType(name='r4.4xlarge', cores=16, memory=122.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r4.8xlarge': InstanceType(name='r4.8xlarge', cores=32, memory=244.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r4.large': InstanceType(name='r4.large', cores=2, memory=15.25, disks=0, disk_capacity=0, architecture='amd64'),
    'r4.xlarge': InstanceType(name='r4.xlarge', cores=4, memory=30.5, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.12xlarge': InstanceType(name='r5.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.16xlarge': InstanceType(name='r5.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.24xlarge': InstanceType(name='r5.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.2xlarge': InstanceType(name='r5.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.4xlarge': InstanceType(name='r5.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.8xlarge': InstanceType(name='r5.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.large': InstanceType(name='r5.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.metal': InstanceType(name='r5.metal', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5.xlarge': InstanceType(name='r5.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.12xlarge': InstanceType(name='r5a.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.16xlarge': InstanceType(name='r5a.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.24xlarge': InstanceType(name='r5a.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.2xlarge': InstanceType(name='r5a.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.4xlarge': InstanceType(name='r5a.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.8xlarge': InstanceType(name='r5a.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.large': InstanceType(name='r5a.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5a.xlarge': InstanceType(name='r5a.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5ad.12xlarge': InstanceType(name='r5ad.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'r5ad.16xlarge': InstanceType(name='r5ad.16xlarge', cores=64, memory=512.0, disks=4.0, disk_capacity=600.0, architecture='amd64'),
    'r5ad.24xlarge': InstanceType(name='r5ad.24xlarge', cores=96, memory=768.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'r5ad.2xlarge': InstanceType(name='r5ad.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'r5ad.4xlarge': InstanceType(name='r5ad.4xlarge', cores=16, memory=128.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'r5ad.8xlarge': InstanceType(name='r5ad.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'r5ad.large': InstanceType(name='r5ad.large', cores=2, memory=16.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'r5ad.xlarge': InstanceType(name='r5ad.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'r5b.12xlarge': InstanceType(name='r5b.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.16xlarge': InstanceType(name='r5b.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.24xlarge': InstanceType(name='r5b.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.2xlarge': InstanceType(name='r5b.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.4xlarge': InstanceType(name='r5b.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.8xlarge': InstanceType(name='r5b.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.large': InstanceType(name='r5b.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.metal': InstanceType(name='r5b.metal', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5b.xlarge': InstanceType(name='r5b.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5d.12xlarge': InstanceType(name='r5d.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'r5d.16xlarge': InstanceType(name='r5d.16xlarge', cores=64, memory=512.0, disks=4.0, disk_capacity=600.0, architecture='amd64'),
    'r5d.24xlarge': InstanceType(name='r5d.24xlarge', cores=96, memory=768.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'r5d.2xlarge': InstanceType(name='r5d.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'r5d.4xlarge': InstanceType(name='r5d.4xlarge', cores=16, memory=128.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'r5d.8xlarge': InstanceType(name='r5d.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'r5d.large': InstanceType(name='r5d.large', cores=2, memory=16.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'r5d.metal': InstanceType(name='r5d.metal', cores=96, memory=768.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'r5d.xlarge': InstanceType(name='r5d.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'r5dn.12xlarge': InstanceType(name='r5dn.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'r5dn.16xlarge': InstanceType(name='r5dn.16xlarge', cores=64, memory=512.0, disks=4.0, disk_capacity=600.0, architecture='amd64'),
    'r5dn.24xlarge': InstanceType(name='r5dn.24xlarge', cores=96, memory=768.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'r5dn.2xlarge': InstanceType(name='r5dn.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'r5dn.4xlarge': InstanceType(name='r5dn.4xlarge', cores=16, memory=128.0, disks=2.0, disk_capacity=300.0, architecture='amd64'),
    'r5dn.8xlarge': InstanceType(name='r5dn.8xlarge', cores=32, memory=256.0, disks=2.0, disk_capacity=600.0, architecture='amd64'),
    'r5dn.large': InstanceType(name='r5dn.large', cores=2, memory=16.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'r5dn.metal': InstanceType(name='r5dn.metal', cores=96, memory=768.0, disks=4.0, disk_capacity=900.0, architecture='amd64'),
    'r5dn.xlarge': InstanceType(name='r5dn.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
    'r5n.12xlarge': InstanceType(name='r5n.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.16xlarge': InstanceType(name='r5n.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.24xlarge': InstanceType(name='r5n.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.2xlarge': InstanceType(name='r5n.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.4xlarge': InstanceType(name='r5n.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.8xlarge': InstanceType(name='r5n.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.large': InstanceType(name='r5n.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.metal': InstanceType(name='r5n.metal', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r5n.xlarge': InstanceType(name='r5n.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.12xlarge': InstanceType(name='r6a.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.16xlarge': InstanceType(name='r6a.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.24xlarge': InstanceType(name='r6a.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.2xlarge': InstanceType(name='r6a.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.32xlarge': InstanceType(name='r6a.32xlarge', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.48xlarge': InstanceType(name='r6a.48xlarge', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.4xlarge': InstanceType(name='r6a.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.8xlarge': InstanceType(name='r6a.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.large': InstanceType(name='r6a.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.metal': InstanceType(name='r6a.metal', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6a.xlarge': InstanceType(name='r6a.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6g.12xlarge': InstanceType(name='r6g.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.16xlarge': InstanceType(name='r6g.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.2xlarge': InstanceType(name='r6g.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.4xlarge': InstanceType(name='r6g.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.8xlarge': InstanceType(name='r6g.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.large': InstanceType(name='r6g.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.medium': InstanceType(name='r6g.medium', cores=1, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.metal': InstanceType(name='r6g.metal', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6g.xlarge': InstanceType(name='r6g.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r6gd.12xlarge': InstanceType(name='r6gd.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'r6gd.16xlarge': InstanceType(name='r6gd.16xlarge', cores=64, memory=512.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'r6gd.2xlarge': InstanceType(name='r6gd.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=475.0, architecture='arm64'),
    'r6gd.4xlarge': InstanceType(name='r6gd.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'r6gd.8xlarge': InstanceType(name='r6gd.8xlarge', cores=32, memory=256.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'r6gd.large': InstanceType(name='r6gd.large', cores=2, memory=16.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'r6gd.medium': InstanceType(name='r6gd.medium', cores=1, memory=8.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'r6gd.metal': InstanceType(name='r6gd.metal', cores=64, memory=512.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'r6gd.xlarge': InstanceType(name='r6gd.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'r6i.12xlarge': InstanceType(name='r6i.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.16xlarge': InstanceType(name='r6i.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.24xlarge': InstanceType(name='r6i.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.2xlarge': InstanceType(name='r6i.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.32xlarge': InstanceType(name='r6i.32xlarge', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.4xlarge': InstanceType(name='r6i.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.8xlarge': InstanceType(name='r6i.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.large': InstanceType(name='r6i.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.metal': InstanceType(name='r6i.metal', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6i.xlarge': InstanceType(name='r6i.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6id.12xlarge': InstanceType(name='r6id.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'r6id.16xlarge': InstanceType(name='r6id.16xlarge', cores=64, memory=512.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'r6id.24xlarge': InstanceType(name='r6id.24xlarge', cores=96, memory=768.0, disks=4.0, disk_capacity=1425.0, architecture='amd64'),
    'r6id.2xlarge': InstanceType(name='r6id.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=474.0, architecture='amd64'),
    'r6id.32xlarge': InstanceType(name='r6id.32xlarge', cores=128, memory=1024.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'r6id.4xlarge': InstanceType(name='r6id.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'r6id.8xlarge': InstanceType(name='r6id.8xlarge', cores=32, memory=256.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'r6id.large': InstanceType(name='r6id.large', cores=2, memory=16.0, disks=1.0, disk_capacity=118.0, architecture='amd64'),
    'r6id.metal': InstanceType(name='r6id.metal', cores=128, memory=1024.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'r6id.xlarge': InstanceType(name='r6id.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=237.0, architecture='amd64'),
    'r6idn.12xlarge': InstanceType(name='r6idn.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'r6idn.16xlarge': InstanceType(name='r6idn.16xlarge', cores=64, memory=512.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'r6idn.24xlarge': InstanceType(name='r6idn.24xlarge', cores=96, memory=768.0, disks=4.0, disk_capacity=1425.0, architecture='amd64'),
    'r6idn.2xlarge': InstanceType(name='r6idn.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=474.0, architecture='amd64'),
    'r6idn.32xlarge': InstanceType(name='r6idn.32xlarge', cores=128, memory=1024.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'r6idn.4xlarge': InstanceType(name='r6idn.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'r6idn.8xlarge': InstanceType(name='r6idn.8xlarge', cores=32, memory=256.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'r6idn.large': InstanceType(name='r6idn.large', cores=2, memory=16.0, disks=1.0, disk_capacity=118.0, architecture='amd64'),
    'r6idn.metal': InstanceType(name='r6idn.metal', cores=128, memory=1024.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'r6idn.xlarge': InstanceType(name='r6idn.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=237.0, architecture='amd64'),
    'r6in.12xlarge': InstanceType(name='r6in.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.16xlarge': InstanceType(name='r6in.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.24xlarge': InstanceType(name='r6in.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.2xlarge': InstanceType(name='r6in.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.32xlarge': InstanceType(name='r6in.32xlarge', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.4xlarge': InstanceType(name='r6in.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.8xlarge': InstanceType(name='r6in.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.large': InstanceType(name='r6in.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.metal': InstanceType(name='r6in.metal', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r6in.xlarge': InstanceType(name='r6in.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.12xlarge': InstanceType(name='r7a.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.16xlarge': InstanceType(name='r7a.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.24xlarge': InstanceType(name='r7a.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.2xlarge': InstanceType(name='r7a.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.32xlarge': InstanceType(name='r7a.32xlarge', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.48xlarge': InstanceType(name='r7a.48xlarge', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.4xlarge': InstanceType(name='r7a.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.8xlarge': InstanceType(name='r7a.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.large': InstanceType(name='r7a.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.medium': InstanceType(name='r7a.medium', cores=1, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.metal-48xl': InstanceType(name='r7a.metal-48xl', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7a.xlarge': InstanceType(name='r7a.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7g.12xlarge': InstanceType(name='r7g.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.16xlarge': InstanceType(name='r7g.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.2xlarge': InstanceType(name='r7g.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.4xlarge': InstanceType(name='r7g.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.8xlarge': InstanceType(name='r7g.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.large': InstanceType(name='r7g.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.medium': InstanceType(name='r7g.medium', cores=1, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.metal': InstanceType(name='r7g.metal', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7g.xlarge': InstanceType(name='r7g.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r7gd.12xlarge': InstanceType(name='r7gd.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'r7gd.16xlarge': InstanceType(name='r7gd.16xlarge', cores=64, memory=512.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'r7gd.2xlarge': InstanceType(name='r7gd.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=475.0, architecture='arm64'),
    'r7gd.4xlarge': InstanceType(name='r7gd.4xlarge', cores=16, memory=128.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'r7gd.8xlarge': InstanceType(name='r7gd.8xlarge', cores=32, memory=256.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'r7gd.large': InstanceType(name='r7gd.large', cores=2, memory=16.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'r7gd.medium': InstanceType(name='r7gd.medium', cores=1, memory=8.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'r7gd.metal': InstanceType(name='r7gd.metal', cores=64, memory=512.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'r7gd.xlarge': InstanceType(name='r7gd.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'r7i.12xlarge': InstanceType(name='r7i.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.16xlarge': InstanceType(name='r7i.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.24xlarge': InstanceType(name='r7i.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.2xlarge': InstanceType(name='r7i.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.48xlarge': InstanceType(name='r7i.48xlarge', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.4xlarge': InstanceType(name='r7i.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.8xlarge': InstanceType(name='r7i.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.large': InstanceType(name='r7i.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.metal-24xl': InstanceType(name='r7i.metal-24xl', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.metal-48xl': InstanceType(name='r7i.metal-48xl', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7i.xlarge': InstanceType(name='r7i.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.12xlarge': InstanceType(name='r7iz.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.16xlarge': InstanceType(name='r7iz.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.2xlarge': InstanceType(name='r7iz.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.32xlarge': InstanceType(name='r7iz.32xlarge', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.4xlarge': InstanceType(name='r7iz.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.8xlarge': InstanceType(name='r7iz.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.large': InstanceType(name='r7iz.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.metal-16xl': InstanceType(name='r7iz.metal-16xl', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.metal-32xl': InstanceType(name='r7iz.metal-32xl', cores=128, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r7iz.xlarge': InstanceType(name='r7iz.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    'r8g.12xlarge': InstanceType(name='r8g.12xlarge', cores=48, memory=384.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.16xlarge': InstanceType(name='r8g.16xlarge', cores=64, memory=512.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.24xlarge': InstanceType(name='r8g.24xlarge', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.2xlarge': InstanceType(name='r8g.2xlarge', cores=8, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.48xlarge': InstanceType(name='r8g.48xlarge', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.4xlarge': InstanceType(name='r8g.4xlarge', cores=16, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.8xlarge': InstanceType(name='r8g.8xlarge', cores=32, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.large': InstanceType(name='r8g.large', cores=2, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.medium': InstanceType(name='r8g.medium', cores=1, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.metal-24xl': InstanceType(name='r8g.metal-24xl', cores=96, memory=768.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.metal-48xl': InstanceType(name='r8g.metal-48xl', cores=192, memory=1536.0, disks=0, disk_capacity=0, architecture='arm64'),
    'r8g.xlarge': InstanceType(name='r8g.xlarge', cores=4, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    't1.micro': InstanceType(name='t1.micro', cores=1, memory=0.613, disks=0, disk_capacity=0, architecture='amd64'),
    't2.2xlarge': InstanceType(name='t2.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    't2.large': InstanceType(name='t2.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    't2.medium': InstanceType(name='t2.medium', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    't2.micro': InstanceType(name='t2.micro', cores=1, memory=1.0, disks=0, disk_capacity=0, architecture='amd64'),
    't2.nano': InstanceType(name='t2.nano', cores=1, memory=0.5, disks=0, disk_capacity=0, architecture='amd64'),
    't2.small': InstanceType(name='t2.small', cores=1, memory=2.0, disks=0, disk_capacity=0, architecture='amd64'),
    't2.xlarge': InstanceType(name='t2.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3.2xlarge': InstanceType(name='t3.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3.large': InstanceType(name='t3.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3.medium': InstanceType(name='t3.medium', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3.micro': InstanceType(name='t3.micro', cores=2, memory=1.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3.nano': InstanceType(name='t3.nano', cores=2, memory=0.5, disks=0, disk_capacity=0, architecture='amd64'),
    't3.small': InstanceType(name='t3.small', cores=2, memory=2.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3.xlarge': InstanceType(name='t3.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.2xlarge': InstanceType(name='t3a.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.large': InstanceType(name='t3a.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.medium': InstanceType(name='t3a.medium', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.micro': InstanceType(name='t3a.micro', cores=2, memory=1.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.nano': InstanceType(name='t3a.nano', cores=2, memory=0.5, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.small': InstanceType(name='t3a.small', cores=2, memory=2.0, disks=0, disk_capacity=0, architecture='amd64'),
    't3a.xlarge': InstanceType(name='t3a.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='amd64'),
    't4g.2xlarge': InstanceType(name='t4g.2xlarge', cores=8, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    't4g.large': InstanceType(name='t4g.large', cores=2, memory=8.0, disks=0, disk_capacity=0, architecture='arm64'),
    't4g.medium': InstanceType(name='t4g.medium', cores=2, memory=4.0, disks=0, disk_capacity=0, architecture='arm64'),
    't4g.micro': InstanceType(name='t4g.micro', cores=2, memory=1.0, disks=0, disk_capacity=0, architecture='arm64'),
    't4g.nano': InstanceType(name='t4g.nano', cores=2, memory=0.5, disks=0, disk_capacity=0, architecture='arm64'),
    't4g.small': InstanceType(name='t4g.small', cores=2, memory=2.0, disks=0, disk_capacity=0, architecture='arm64'),
    't4g.xlarge': InstanceType(name='t4g.xlarge', cores=4, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'trn1.2xlarge': InstanceType(name='trn1.2xlarge', cores=8, memory=32.0, disks=1.0, disk_capacity=475.0, architecture='amd64'),
    'trn1.32xlarge': InstanceType(name='trn1.32xlarge', cores=128, memory=512.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'trn1n.32xlarge': InstanceType(name='trn1n.32xlarge', cores=128, memory=512.0, disks=4.0, disk_capacity=1900.0, architecture='amd64'),
    'trn2.48xlarge': InstanceType(name='trn2.48xlarge', cores=192, memory=2048.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-12tb1.112xlarge': InstanceType(name='u-12tb1.112xlarge', cores=448, memory=12288.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-18tb1.112xlarge': InstanceType(name='u-18tb1.112xlarge', cores=448, memory=18432.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-24tb1.112xlarge': InstanceType(name='u-24tb1.112xlarge', cores=448, memory=24576.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-3tb1.56xlarge': InstanceType(name='u-3tb1.56xlarge', cores=224, memory=3072.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-6tb1.112xlarge': InstanceType(name='u-6tb1.112xlarge', cores=448, memory=6144.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-6tb1.56xlarge': InstanceType(name='u-6tb1.56xlarge', cores=224, memory=6144.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u-9tb1.112xlarge': InstanceType(name='u-9tb1.112xlarge', cores=448, memory=9216.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u7i-12tb.224xlarge': InstanceType(name='u7i-12tb.224xlarge', cores=896, memory=12288.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u7i-6tb.112xlarge': InstanceType(name='u7i-6tb.112xlarge', cores=448, memory=6144.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u7i-8tb.112xlarge': InstanceType(name='u7i-8tb.112xlarge', cores=448, memory=8192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u7in-16tb.224xlarge': InstanceType(name='u7in-16tb.224xlarge', cores=896, memory=16384.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u7in-24tb.224xlarge': InstanceType(name='u7in-24tb.224xlarge', cores=896, memory=24576.0, disks=0, disk_capacity=0, architecture='amd64'),
    'u7in-32tb.224xlarge': InstanceType(name='u7in-32tb.224xlarge', cores=896, memory=32768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'vt1.24xlarge': InstanceType(name='vt1.24xlarge', cores=96, memory=192.0, disks=0, disk_capacity=0, architecture='amd64'),
    'vt1.3xlarge': InstanceType(name='vt1.3xlarge', cores=12, memory=24.0, disks=0, disk_capacity=0, architecture='amd64'),
    'vt1.6xlarge': InstanceType(name='vt1.6xlarge', cores=24, memory=48.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x1.16xlarge': InstanceType(name='x1.16xlarge', cores=64, memory=976.0, disks=1.0, disk_capacity=1920.0, architecture='amd64'),
    'x1.32xlarge': InstanceType(name='x1.32xlarge', cores=128, memory=1952.0, disks=2.0, disk_capacity=1920.0, architecture='amd64'),
    'x1e.16xlarge': InstanceType(name='x1e.16xlarge', cores=64, memory=1952.0, disks=1.0, disk_capacity=1920.0, architecture='amd64'),
    'x1e.2xlarge': InstanceType(name='x1e.2xlarge', cores=8, memory=244.0, disks=1.0, disk_capacity=240.0, architecture='amd64'),
    'x1e.32xlarge': InstanceType(name='x1e.32xlarge', cores=128, memory=3904.0, disks=2.0, disk_capacity=1920.0, architecture='amd64'),
    'x1e.4xlarge': InstanceType(name='x1e.4xlarge', cores=16, memory=488.0, disks=1.0, disk_capacity=480.0, architecture='amd64'),
    'x1e.8xlarge': InstanceType(name='x1e.8xlarge', cores=32, memory=976.0, disks=1.0, disk_capacity=960.0, architecture='amd64'),
    'x1e.xlarge': InstanceType(name='x1e.xlarge', cores=4, memory=122.0, disks=1.0, disk_capacity=120.0, architecture='amd64'),
    'x2gd.12xlarge': InstanceType(name='x2gd.12xlarge', cores=48, memory=768.0, disks=2.0, disk_capacity=1425.0, architecture='arm64'),
    'x2gd.16xlarge': InstanceType(name='x2gd.16xlarge', cores=64, memory=1024.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'x2gd.2xlarge': InstanceType(name='x2gd.2xlarge', cores=8, memory=128.0, disks=1.0, disk_capacity=474.0, architecture='arm64'),
    'x2gd.4xlarge': InstanceType(name='x2gd.4xlarge', cores=16, memory=256.0, disks=1.0, disk_capacity=950.0, architecture='arm64'),
    'x2gd.8xlarge': InstanceType(name='x2gd.8xlarge', cores=32, memory=512.0, disks=1.0, disk_capacity=1900.0, architecture='arm64'),
    'x2gd.large': InstanceType(name='x2gd.large', cores=2, memory=32.0, disks=1.0, disk_capacity=118.0, architecture='arm64'),
    'x2gd.medium': InstanceType(name='x2gd.medium', cores=1, memory=16.0, disks=1.0, disk_capacity=59.0, architecture='arm64'),
    'x2gd.metal': InstanceType(name='x2gd.metal', cores=64, memory=1024.0, disks=2.0, disk_capacity=1900.0, architecture='arm64'),
    'x2gd.xlarge': InstanceType(name='x2gd.xlarge', cores=4, memory=64.0, disks=1.0, disk_capacity=237.0, architecture='arm64'),
    'x2idn.16xlarge': InstanceType(name='x2idn.16xlarge', cores=64, memory=1024.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'x2idn.24xlarge': InstanceType(name='x2idn.24xlarge', cores=96, memory=1536.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'x2idn.32xlarge': InstanceType(name='x2idn.32xlarge', cores=128, memory=2048.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'x2idn.metal': InstanceType(name='x2idn.metal', cores=128, memory=2048.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'x2iedn.16xlarge': InstanceType(name='x2iedn.16xlarge', cores=64, memory=2048.0, disks=1.0, disk_capacity=1900.0, architecture='amd64'),
    'x2iedn.24xlarge': InstanceType(name='x2iedn.24xlarge', cores=96, memory=3072.0, disks=2.0, disk_capacity=1425.0, architecture='amd64'),
    'x2iedn.2xlarge': InstanceType(name='x2iedn.2xlarge', cores=8, memory=256.0, disks=1.0, disk_capacity=237.0, architecture='amd64'),
    'x2iedn.32xlarge': InstanceType(name='x2iedn.32xlarge', cores=128, memory=4096.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'x2iedn.4xlarge': InstanceType(name='x2iedn.4xlarge', cores=16, memory=512.0, disks=1.0, disk_capacity=475.0, architecture='amd64'),
    'x2iedn.8xlarge': InstanceType(name='x2iedn.8xlarge', cores=32, memory=1024.0, disks=1.0, disk_capacity=950.0, architecture='amd64'),
    'x2iedn.metal': InstanceType(name='x2iedn.metal', cores=128, memory=4096.0, disks=2.0, disk_capacity=1900.0, architecture='amd64'),
    'x2iedn.xlarge': InstanceType(name='x2iedn.xlarge', cores=4, memory=128.0, disks=1.0, disk_capacity=118.0, architecture='amd64'),
    'x2iezn.12xlarge': InstanceType(name='x2iezn.12xlarge', cores=48, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x2iezn.2xlarge': InstanceType(name='x2iezn.2xlarge', cores=8, memory=256.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x2iezn.4xlarge': InstanceType(name='x2iezn.4xlarge', cores=16, memory=512.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x2iezn.6xlarge': InstanceType(name='x2iezn.6xlarge', cores=24, memory=768.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x2iezn.8xlarge': InstanceType(name='x2iezn.8xlarge', cores=32, memory=1024.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x2iezn.metal': InstanceType(name='x2iezn.metal', cores=48, memory=1536.0, disks=0, disk_capacity=0, architecture='amd64'),
    'x8g.12xlarge': InstanceType(name='x8g.12xlarge', cores=48, memory=768.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.16xlarge': InstanceType(name='x8g.16xlarge', cores=64, memory=1024.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.24xlarge': InstanceType(name='x8g.24xlarge', cores=96, memory=1536.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.2xlarge': InstanceType(name='x8g.2xlarge', cores=8, memory=128.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.48xlarge': InstanceType(name='x8g.48xlarge', cores=192, memory=3072.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.4xlarge': InstanceType(name='x8g.4xlarge', cores=16, memory=256.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.8xlarge': InstanceType(name='x8g.8xlarge', cores=32, memory=512.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.large': InstanceType(name='x8g.large', cores=2, memory=32.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.medium': InstanceType(name='x8g.medium', cores=1, memory=16.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.metal-24xl': InstanceType(name='x8g.metal-24xl', cores=96, memory=1536.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.metal-48xl': InstanceType(name='x8g.metal-48xl', cores=192, memory=3072.0, disks=0, disk_capacity=0, architecture='arm64'),
    'x8g.xlarge': InstanceType(name='x8g.xlarge', cores=4, memory=64.0, disks=0, disk_capacity=0, architecture='arm64'),
    'z1d.12xlarge': InstanceType(name='z1d.12xlarge', cores=48, memory=384.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'z1d.2xlarge': InstanceType(name='z1d.2xlarge', cores=8, memory=64.0, disks=1.0, disk_capacity=300.0, architecture='amd64'),
    'z1d.3xlarge': InstanceType(name='z1d.3xlarge', cores=12, memory=96.0, disks=1.0, disk_capacity=450.0, architecture='amd64'),
    'z1d.6xlarge': InstanceType(name='z1d.6xlarge', cores=24, memory=192.0, disks=1.0, disk_capacity=900.0, architecture='amd64'),
    'z1d.large': InstanceType(name='z1d.large', cores=2, memory=16.0, disks=1.0, disk_capacity=75.0, architecture='amd64'),
    'z1d.metal': InstanceType(name='z1d.metal', cores=48, memory=384.0, disks=2.0, disk_capacity=900.0, architecture='amd64'),
    'z1d.xlarge': InstanceType(name='z1d.xlarge', cores=4, memory=32.0, disks=1.0, disk_capacity=150.0, architecture='amd64'),
}

regionDict = {
              'us-west-1': ['c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p5.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'us-west-2': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7a.12xlarge', 'c7a.16xlarge', 'c7a.24xlarge', 'c7a.2xlarge', 'c7a.32xlarge', 'c7a.48xlarge', 'c7a.4xlarge', 'c7a.8xlarge', 'c7a.large', 'c7a.medium', 'c7a.metal-48xl', 'c7a.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7gn.12xlarge', 'c7gn.16xlarge', 'c7gn.2xlarge', 'c7gn.4xlarge', 'c7gn.8xlarge', 'c7gn.large', 'c7gn.medium', 'c7gn.metal', 'c7gn.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'c8g.12xlarge', 'c8g.16xlarge', 'c8g.24xlarge', 'c8g.2xlarge', 'c8g.48xlarge', 'c8g.4xlarge', 'c8g.8xlarge', 'c8g.large', 'c8g.medium', 'c8g.metal-24xl', 'c8g.metal-48xl', 'c8g.xlarge', 'cr1.8xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'dl1.24xlarge', 'dl2q.24xlarge', 'f1.16xlarge', 'f1.2xlarge', 'f1.4xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g5g.16xlarge', 'g5g.2xlarge', 'g5g.4xlarge', 'g5g.8xlarge', 'g5g.metal', 'g5g.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'g6e.12xlarge', 'g6e.16xlarge', 'g6e.24xlarge', 'g6e.2xlarge', 'g6e.48xlarge', 'g6e.4xlarge', 'g6e.8xlarge', 'g6e.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'h1.16xlarge', 'h1.2xlarge', 'h1.4xlarge', 'h1.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4g.16xlarge', 'i4g.2xlarge', 'i4g.4xlarge', 'i4g.8xlarge', 'i4g.large', 'i4g.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'i7ie.12xlarge', 'i7ie.18xlarge', 'i7ie.24xlarge', 'i7ie.2xlarge', 'i7ie.3xlarge', 'i7ie.48xlarge', 'i7ie.6xlarge', 'i7ie.large', 'i7ie.xlarge', 'i8g.12xlarge', 'i8g.16xlarge', 'i8g.24xlarge', 'i8g.2xlarge', 'i8g.4xlarge', 'i8g.8xlarge', 'i8g.large', 'i8g.metal-24xl', 'i8g.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7a.12xlarge', 'm7a.16xlarge', 'm7a.24xlarge', 'm7a.2xlarge', 'm7a.32xlarge', 'm7a.48xlarge', 'm7a.4xlarge', 'm7a.8xlarge', 'm7a.large', 'm7a.medium', 'm7a.metal-48xl', 'm7a.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'm8g.12xlarge', 'm8g.16xlarge', 'm8g.24xlarge', 'm8g.2xlarge', 'm8g.48xlarge', 'm8g.4xlarge', 'm8g.8xlarge', 'm8g.large', 'm8g.medium', 'm8g.metal-24xl', 'm8g.metal-48xl', 'm8g.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p3dn.24xlarge', 'p4d.24xlarge', 'p4de.24xlarge', 'p5.48xlarge', 'p5en.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7a.12xlarge', 'r7a.16xlarge', 'r7a.24xlarge', 'r7a.2xlarge', 'r7a.32xlarge', 'r7a.48xlarge', 'r7a.4xlarge', 'r7a.8xlarge', 'r7a.large', 'r7a.medium', 'r7a.metal-48xl', 'r7a.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r7iz.12xlarge', 'r7iz.16xlarge', 'r7iz.2xlarge', 'r7iz.32xlarge', 'r7iz.4xlarge', 'r7iz.8xlarge', 'r7iz.large', 'r7iz.metal-16xl', 'r7iz.metal-32xl', 'r7iz.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'trn1.2xlarge', 'trn1.32xlarge', 'trn1n.32xlarge', 'u-12tb1.112xlarge', 'u-18tb1.112xlarge', 'u-24tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'u7i-12tb.224xlarge', 'u7i-6tb.112xlarge', 'u7in-16tb.224xlarge', 'u7in-24tb.224xlarge', 'u7in-32tb.224xlarge', 'vt1.24xlarge', 'vt1.3xlarge', 'vt1.6xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2gd.12xlarge', 'x2gd.16xlarge', 'x2gd.2xlarge', 'x2gd.4xlarge', 'x2gd.8xlarge', 'x2gd.large', 'x2gd.medium', 'x2gd.metal', 'x2gd.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'x2iezn.12xlarge', 'x2iezn.2xlarge', 'x2iezn.4xlarge', 'x2iezn.6xlarge', 'x2iezn.8xlarge', 'x2iezn.metal', 'x8g.12xlarge', 'x8g.16xlarge', 'x8g.24xlarge', 'x8g.2xlarge', 'x8g.48xlarge', 'x8g.4xlarge', 'x8g.8xlarge', 'x8g.large', 'x8g.medium', 'x8g.metal-24xl', 'x8g.metal-48xl', 'x8g.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'us-east-1': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7a.12xlarge', 'c7a.16xlarge', 'c7a.24xlarge', 'c7a.2xlarge', 'c7a.32xlarge', 'c7a.48xlarge', 'c7a.4xlarge', 'c7a.8xlarge', 'c7a.large', 'c7a.medium', 'c7a.metal-48xl', 'c7a.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7gn.12xlarge', 'c7gn.16xlarge', 'c7gn.2xlarge', 'c7gn.4xlarge', 'c7gn.8xlarge', 'c7gn.large', 'c7gn.medium', 'c7gn.metal', 'c7gn.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'c8g.12xlarge', 'c8g.16xlarge', 'c8g.24xlarge', 'c8g.2xlarge', 'c8g.48xlarge', 'c8g.4xlarge', 'c8g.8xlarge', 'c8g.large', 'c8g.medium', 'c8g.metal-24xl', 'c8g.metal-48xl', 'c8g.xlarge', 'cr1.8xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'dl1.24xlarge', 'f1.16xlarge', 'f1.2xlarge', 'f1.4xlarge', 'f2.12xlarge', 'f2.48xlarge', 'f2.6xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g5g.16xlarge', 'g5g.2xlarge', 'g5g.4xlarge', 'g5g.8xlarge', 'g5g.metal', 'g5g.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'g6e.12xlarge', 'g6e.16xlarge', 'g6e.24xlarge', 'g6e.2xlarge', 'g6e.48xlarge', 'g6e.4xlarge', 'g6e.8xlarge', 'g6e.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'h1.16xlarge', 'h1.2xlarge', 'h1.4xlarge', 'h1.8xlarge', 'hpc7g.16xlarge', 'hpc7g.4xlarge', 'hpc7g.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4g.16xlarge', 'i4g.2xlarge', 'i4g.4xlarge', 'i4g.8xlarge', 'i4g.large', 'i4g.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'i7ie.12xlarge', 'i7ie.18xlarge', 'i7ie.24xlarge', 'i7ie.2xlarge', 'i7ie.3xlarge', 'i7ie.48xlarge', 'i7ie.6xlarge', 'i7ie.large', 'i7ie.xlarge', 'i8g.12xlarge', 'i8g.16xlarge', 'i8g.24xlarge', 'i8g.2xlarge', 'i8g.4xlarge', 'i8g.8xlarge', 'i8g.large', 'i8g.metal-24xl', 'i8g.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7a.12xlarge', 'm7a.16xlarge', 'm7a.24xlarge', 'm7a.2xlarge', 'm7a.32xlarge', 'm7a.48xlarge', 'm7a.4xlarge', 'm7a.8xlarge', 'm7a.large', 'm7a.medium', 'm7a.metal-48xl', 'm7a.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'm8g.12xlarge', 'm8g.16xlarge', 'm8g.24xlarge', 'm8g.2xlarge', 'm8g.48xlarge', 'm8g.4xlarge', 'm8g.8xlarge', 'm8g.large', 'm8g.medium', 'm8g.metal-24xl', 'm8g.metal-48xl', 'm8g.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p3dn.24xlarge', 'p4d.24xlarge', 'p4de.24xlarge', 'p5.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7a.12xlarge', 'r7a.16xlarge', 'r7a.24xlarge', 'r7a.2xlarge', 'r7a.32xlarge', 'r7a.48xlarge', 'r7a.4xlarge', 'r7a.8xlarge', 'r7a.large', 'r7a.medium', 'r7a.metal-48xl', 'r7a.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r7iz.12xlarge', 'r7iz.16xlarge', 'r7iz.2xlarge', 'r7iz.32xlarge', 'r7iz.4xlarge', 'r7iz.8xlarge', 'r7iz.large', 'r7iz.metal-16xl', 'r7iz.metal-32xl', 'r7iz.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'trn1.2xlarge', 'trn1.32xlarge', 'trn1n.32xlarge', 'u-12tb1.112xlarge', 'u-18tb1.112xlarge', 'u-24tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'u7i-12tb.224xlarge', 'u7i-6tb.112xlarge', 'u7i-8tb.112xlarge', 'u7in-16tb.224xlarge', 'u7in-24tb.224xlarge', 'u7in-32tb.224xlarge', 'vt1.24xlarge', 'vt1.3xlarge', 'vt1.6xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2gd.12xlarge', 'x2gd.16xlarge', 'x2gd.2xlarge', 'x2gd.4xlarge', 'x2gd.8xlarge', 'x2gd.large', 'x2gd.medium', 'x2gd.metal', 'x2gd.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'x2iezn.12xlarge', 'x2iezn.2xlarge', 'x2iezn.4xlarge', 'x2iezn.6xlarge', 'x2iezn.8xlarge', 'x2iezn.metal', 'x8g.12xlarge', 'x8g.16xlarge', 'x8g.24xlarge', 'x8g.2xlarge', 'x8g.48xlarge', 'x8g.4xlarge', 'x8g.8xlarge', 'x8g.large', 'x8g.medium', 'x8g.metal-24xl', 'x8g.metal-48xl', 'x8g.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'us-east-2': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7a.12xlarge', 'c7a.16xlarge', 'c7a.24xlarge', 'c7a.2xlarge', 'c7a.32xlarge', 'c7a.48xlarge', 'c7a.4xlarge', 'c7a.8xlarge', 'c7a.large', 'c7a.medium', 'c7a.metal-48xl', 'c7a.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7gn.12xlarge', 'c7gn.16xlarge', 'c7gn.2xlarge', 'c7gn.4xlarge', 'c7gn.8xlarge', 'c7gn.large', 'c7gn.medium', 'c7gn.metal', 'c7gn.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'c8g.12xlarge', 'c8g.16xlarge', 'c8g.24xlarge', 'c8g.2xlarge', 'c8g.48xlarge', 'c8g.4xlarge', 'c8g.8xlarge', 'c8g.large', 'c8g.medium', 'c8g.metal-24xl', 'c8g.metal-48xl', 'c8g.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'g6e.12xlarge', 'g6e.16xlarge', 'g6e.24xlarge', 'g6e.2xlarge', 'g6e.48xlarge', 'g6e.4xlarge', 'g6e.8xlarge', 'g6e.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'h1.16xlarge', 'h1.2xlarge', 'h1.4xlarge', 'h1.8xlarge', 'hpc6a.48xlarge', 'hpc6id.32xlarge', 'hpc7a.12xlarge', 'hpc7a.24xlarge', 'hpc7a.48xlarge', 'hpc7a.96xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4g.16xlarge', 'i4g.2xlarge', 'i4g.4xlarge', 'i4g.8xlarge', 'i4g.large', 'i4g.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'i7ie.12xlarge', 'i7ie.18xlarge', 'i7ie.24xlarge', 'i7ie.2xlarge', 'i7ie.3xlarge', 'i7ie.48xlarge', 'i7ie.6xlarge', 'i7ie.large', 'i7ie.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7a.12xlarge', 'm7a.16xlarge', 'm7a.24xlarge', 'm7a.2xlarge', 'm7a.32xlarge', 'm7a.48xlarge', 'm7a.4xlarge', 'm7a.8xlarge', 'm7a.large', 'm7a.medium', 'm7a.metal-48xl', 'm7a.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'm8g.12xlarge', 'm8g.16xlarge', 'm8g.24xlarge', 'm8g.2xlarge', 'm8g.48xlarge', 'm8g.4xlarge', 'm8g.8xlarge', 'm8g.large', 'm8g.medium', 'm8g.metal-24xl', 'm8g.metal-48xl', 'm8g.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'p5en.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7a.12xlarge', 'r7a.16xlarge', 'r7a.24xlarge', 'r7a.2xlarge', 'r7a.32xlarge', 'r7a.48xlarge', 'r7a.4xlarge', 'r7a.8xlarge', 'r7a.large', 'r7a.medium', 'r7a.metal-48xl', 'r7a.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r7iz.12xlarge', 'r7iz.16xlarge', 'r7iz.2xlarge', 'r7iz.32xlarge', 'r7iz.4xlarge', 'r7iz.8xlarge', 'r7iz.large', 'r7iz.metal-16xl', 'r7iz.metal-32xl', 'r7iz.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'trn1.2xlarge', 'trn1.32xlarge', 'trn1n.32xlarge', 'trn2.48xlarge', 'u-12tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2gd.12xlarge', 'x2gd.16xlarge', 'x2gd.2xlarge', 'x2gd.4xlarge', 'x2gd.8xlarge', 'x2gd.large', 'x2gd.medium', 'x2gd.metal', 'x2gd.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'us-gov-west-1': ['c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'cc2.8xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'f1.16xlarge', 'f1.2xlarge', 'f1.4xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'hpc6a.48xlarge', 'hpc6id.32xlarge', 'hpc7a.12xlarge', 'hpc7a.24xlarge', 'hpc7a.48xlarge', 'hpc7a.96xlarge', 'hpc7g.16xlarge', 'hpc7g.4xlarge', 'hpc7g.8xlarge', 'hs1.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i3p.16xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p3dn.24xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-24tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'u7in-24tb.224xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge'],
              'ca-central-1': ['c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.metal', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4g.16xlarge', 'i4g.2xlarge', 'i4g.4xlarge', 'i4g.8xlarge', 'i4g.large', 'i4g.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge'],
              'ap-northeast-1': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7a.12xlarge', 'c7a.16xlarge', 'c7a.24xlarge', 'c7a.2xlarge', 'c7a.32xlarge', 'c7a.48xlarge', 'c7a.4xlarge', 'c7a.8xlarge', 'c7a.large', 'c7a.medium', 'c7a.metal-48xl', 'c7a.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7gn.12xlarge', 'c7gn.16xlarge', 'c7gn.2xlarge', 'c7gn.4xlarge', 'c7gn.8xlarge', 'c7gn.large', 'c7gn.medium', 'c7gn.metal', 'c7gn.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'cr1.8xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g5g.16xlarge', 'g5g.2xlarge', 'g5g.4xlarge', 'g5g.8xlarge', 'g5g.metal', 'g5g.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'g6e.12xlarge', 'g6e.16xlarge', 'g6e.24xlarge', 'g6e.2xlarge', 'g6e.48xlarge', 'g6e.4xlarge', 'g6e.8xlarge', 'g6e.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'hpc7g.16xlarge', 'hpc7g.4xlarge', 'hpc7g.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'i7ie.12xlarge', 'i7ie.18xlarge', 'i7ie.24xlarge', 'i7ie.2xlarge', 'i7ie.3xlarge', 'i7ie.48xlarge', 'i7ie.6xlarge', 'i7ie.large', 'i7ie.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7a.12xlarge', 'm7a.16xlarge', 'm7a.24xlarge', 'm7a.2xlarge', 'm7a.32xlarge', 'm7a.48xlarge', 'm7a.4xlarge', 'm7a.8xlarge', 'm7a.large', 'm7a.medium', 'm7a.metal-48xl', 'm7a.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p3dn.24xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'p5en.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7a.12xlarge', 'r7a.16xlarge', 'r7a.24xlarge', 'r7a.2xlarge', 'r7a.32xlarge', 'r7a.48xlarge', 'r7a.4xlarge', 'r7a.8xlarge', 'r7a.large', 'r7a.medium', 'r7a.metal-48xl', 'r7a.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r7iz.12xlarge', 'r7iz.16xlarge', 'r7iz.2xlarge', 'r7iz.32xlarge', 'r7iz.4xlarge', 'r7iz.8xlarge', 'r7iz.large', 'r7iz.metal-16xl', 'r7iz.metal-32xl', 'r7iz.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'vt1.24xlarge', 'vt1.3xlarge', 'vt1.6xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'x2iezn.12xlarge', 'x2iezn.2xlarge', 'x2iezn.4xlarge', 'x2iezn.6xlarge', 'x2iezn.8xlarge', 'x2iezn.metal', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'ap-northeast-2': ['c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g5g.16xlarge', 'g5g.2xlarge', 'g5g.4xlarge', 'g5g.8xlarge', 'g5g.metal', 'g5g.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p4d.24xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-24tb1.112xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'u7i-6tb.112xlarge', 'u7in-16tb.224xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'ap-northeast-3': ['c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge'],
              'ap-southeast-1': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5g.16xlarge', 'g5g.2xlarge', 'g5g.4xlarge', 'g5g.8xlarge', 'g5g.metal', 'g5g.xlarge', 'hpc6a.48xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4g.16xlarge', 'i4g.2xlarge', 'i4g.4xlarge', 'i4g.8xlarge', 'i4g.large', 'i4g.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p4d.24xlarge', 'p4de.24xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'ap-southeast-2': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'f1.16xlarge', 'f1.2xlarge', 'f1.4xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'hpc6a.48xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u7in-16tb.224xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'ap-south-1': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'eu-west-1': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7a.12xlarge', 'c7a.16xlarge', 'c7a.24xlarge', 'c7a.2xlarge', 'c7a.32xlarge', 'c7a.48xlarge', 'c7a.4xlarge', 'c7a.8xlarge', 'c7a.large', 'c7a.medium', 'c7a.metal-48xl', 'c7a.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7gn.12xlarge', 'c7gn.16xlarge', 'c7gn.2xlarge', 'c7gn.4xlarge', 'c7gn.8xlarge', 'c7gn.large', 'c7gn.medium', 'c7gn.metal', 'c7gn.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'c8g.12xlarge', 'c8g.16xlarge', 'c8g.24xlarge', 'c8g.2xlarge', 'c8g.48xlarge', 'c8g.4xlarge', 'c8g.8xlarge', 'c8g.large', 'c8g.medium', 'c8g.metal-24xl', 'c8g.metal-48xl', 'c8g.xlarge', 'cr1.8xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'f1.16xlarge', 'f1.2xlarge', 'f1.4xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'h1.16xlarge', 'h1.2xlarge', 'h1.4xlarge', 'h1.8xlarge', 'hpc7a.12xlarge', 'hpc7a.24xlarge', 'hpc7a.48xlarge', 'hpc7a.96xlarge', 'hpc7g.16xlarge', 'hpc7g.4xlarge', 'hpc7g.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4g.16xlarge', 'i4g.2xlarge', 'i4g.4xlarge', 'i4g.8xlarge', 'i4g.large', 'i4g.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7a.12xlarge', 'm7a.16xlarge', 'm7a.24xlarge', 'm7a.2xlarge', 'm7a.32xlarge', 'm7a.48xlarge', 'm7a.4xlarge', 'm7a.8xlarge', 'm7a.large', 'm7a.medium', 'm7a.metal-48xl', 'm7a.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p3dn.24xlarge', 'p4d.24xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7a.12xlarge', 'r7a.16xlarge', 'r7a.24xlarge', 'r7a.2xlarge', 'r7a.32xlarge', 'r7a.48xlarge', 'r7a.4xlarge', 'r7a.8xlarge', 'r7a.large', 'r7a.medium', 'r7a.metal-48xl', 'r7a.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r7iz.12xlarge', 'r7iz.16xlarge', 'r7iz.2xlarge', 'r7iz.32xlarge', 'r7iz.4xlarge', 'r7iz.8xlarge', 'r7iz.large', 'r7iz.metal-16xl', 'r7iz.metal-32xl', 'r7iz.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-18tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'vt1.24xlarge', 'vt1.3xlarge', 'vt1.6xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2gd.12xlarge', 'x2gd.16xlarge', 'x2gd.2xlarge', 'x2gd.4xlarge', 'x2gd.8xlarge', 'x2gd.large', 'x2gd.medium', 'x2gd.metal', 'x2gd.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'x2iezn.12xlarge', 'x2iezn.2xlarge', 'x2iezn.4xlarge', 'x2iezn.6xlarge', 'x2iezn.8xlarge', 'x2iezn.metal', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'eu-west-2': ['c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'f1.2xlarge', 'f1.4xlarge', 'f2.12xlarge', 'f2.48xlarge', 'f2.6xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'i7ie.12xlarge', 'i7ie.18xlarge', 'i7ie.24xlarge', 'i7ie.2xlarge', 'i7ie.3xlarge', 'i7ie.48xlarge', 'i7ie.6xlarge', 'i7ie.large', 'i7ie.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p5.48xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'eu-west-3': ['c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5d.18xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.metal', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'hpc6id.32xlarge', 'hpc7a.12xlarge', 'hpc7a.24xlarge', 'hpc7a.48xlarge', 'hpc7a.96xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge'],
              'eu-central-1': ['a1.2xlarge', 'a1.4xlarge', 'a1.large', 'a1.medium', 'a1.metal', 'a1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7a.12xlarge', 'c7a.16xlarge', 'c7a.24xlarge', 'c7a.2xlarge', 'c7a.32xlarge', 'c7a.48xlarge', 'c7a.4xlarge', 'c7a.8xlarge', 'c7a.large', 'c7a.medium', 'c7a.metal-48xl', 'c7a.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7gd.12xlarge', 'c7gd.16xlarge', 'c7gd.2xlarge', 'c7gd.4xlarge', 'c7gd.8xlarge', 'c7gd.large', 'c7gd.medium', 'c7gd.metal', 'c7gd.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'c8g.12xlarge', 'c8g.16xlarge', 'c8g.24xlarge', 'c8g.2xlarge', 'c8g.48xlarge', 'c8g.4xlarge', 'c8g.8xlarge', 'c8g.large', 'c8g.medium', 'c8g.metal-24xl', 'c8g.metal-48xl', 'c8g.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge', 'd2.xlarge', 'd3.2xlarge', 'd3.4xlarge', 'd3.8xlarge', 'd3.xlarge', 'd3en.12xlarge', 'd3en.2xlarge', 'd3en.4xlarge', 'd3en.6xlarge', 'd3en.8xlarge', 'd3en.xlarge', 'dl2q.24xlarge', 'f1.2xlarge', 'f1.4xlarge', 'g2.2xlarge', 'g2.8xlarge', 'g3.16xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3s.xlarge', 'g4ad.16xlarge', 'g4ad.2xlarge', 'g4ad.4xlarge', 'g4ad.8xlarge', 'g4ad.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g5g.16xlarge', 'g5g.2xlarge', 'g5g.4xlarge', 'g5g.8xlarge', 'g5g.metal', 'g5g.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'g6e.12xlarge', 'g6e.16xlarge', 'g6e.24xlarge', 'g6e.2xlarge', 'g6e.48xlarge', 'g6e.4xlarge', 'g6e.8xlarge', 'g6e.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge', 'i2.xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'i7ie.12xlarge', 'i7ie.18xlarge', 'i7ie.24xlarge', 'i7ie.2xlarge', 'i7ie.3xlarge', 'i7ie.48xlarge', 'i7ie.6xlarge', 'i7ie.large', 'i7ie.xlarge', 'i8g.12xlarge', 'i8g.16xlarge', 'i8g.24xlarge', 'i8g.2xlarge', 'i8g.4xlarge', 'i8g.8xlarge', 'i8g.large', 'i8g.metal-24xl', 'i8g.xlarge', 'im4gn.16xlarge', 'im4gn.2xlarge', 'im4gn.4xlarge', 'im4gn.8xlarge', 'im4gn.large', 'im4gn.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'is4gen.2xlarge', 'is4gen.4xlarge', 'is4gen.8xlarge', 'is4gen.large', 'is4gen.medium', 'is4gen.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5dn.12xlarge', 'm5dn.16xlarge', 'm5dn.24xlarge', 'm5dn.2xlarge', 'm5dn.4xlarge', 'm5dn.8xlarge', 'm5dn.large', 'm5dn.metal', 'm5dn.xlarge', 'm5n.12xlarge', 'm5n.16xlarge', 'm5n.24xlarge', 'm5n.2xlarge', 'm5n.4xlarge', 'm5n.8xlarge', 'm5n.large', 'm5n.metal', 'm5n.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm6idn.12xlarge', 'm6idn.16xlarge', 'm6idn.24xlarge', 'm6idn.2xlarge', 'm6idn.32xlarge', 'm6idn.4xlarge', 'm6idn.8xlarge', 'm6idn.large', 'm6idn.metal', 'm6idn.xlarge', 'm6in.12xlarge', 'm6in.16xlarge', 'm6in.24xlarge', 'm6in.2xlarge', 'm6in.32xlarge', 'm6in.4xlarge', 'm6in.8xlarge', 'm6in.large', 'm6in.metal', 'm6in.xlarge', 'm7a.12xlarge', 'm7a.16xlarge', 'm7a.24xlarge', 'm7a.2xlarge', 'm7a.32xlarge', 'm7a.48xlarge', 'm7a.4xlarge', 'm7a.8xlarge', 'm7a.large', 'm7a.medium', 'm7a.metal-48xl', 'm7a.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'm8g.12xlarge', 'm8g.16xlarge', 'm8g.24xlarge', 'm8g.2xlarge', 'm8g.48xlarge', 'm8g.4xlarge', 'm8g.8xlarge', 'm8g.large', 'm8g.medium', 'm8g.metal-24xl', 'm8g.metal-48xl', 'm8g.xlarge', 'p2.16xlarge', 'p2.8xlarge', 'p2.xlarge', 'p3.16xlarge', 'p3.2xlarge', 'p3.8xlarge', 'p4d.24xlarge', 'p4de.24xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5dn.12xlarge', 'r5dn.16xlarge', 'r5dn.24xlarge', 'r5dn.2xlarge', 'r5dn.4xlarge', 'r5dn.8xlarge', 'r5dn.large', 'r5dn.metal', 'r5dn.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6a.12xlarge', 'r6a.16xlarge', 'r6a.24xlarge', 'r6a.2xlarge', 'r6a.32xlarge', 'r6a.48xlarge', 'r6a.4xlarge', 'r6a.8xlarge', 'r6a.large', 'r6a.metal', 'r6a.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r6id.12xlarge', 'r6id.16xlarge', 'r6id.24xlarge', 'r6id.2xlarge', 'r6id.32xlarge', 'r6id.4xlarge', 'r6id.8xlarge', 'r6id.large', 'r6id.metal', 'r6id.xlarge', 'r6idn.12xlarge', 'r6idn.16xlarge', 'r6idn.24xlarge', 'r6idn.2xlarge', 'r6idn.32xlarge', 'r6idn.4xlarge', 'r6idn.8xlarge', 'r6idn.large', 'r6idn.metal', 'r6idn.xlarge', 'r6in.12xlarge', 'r6in.16xlarge', 'r6in.24xlarge', 'r6in.2xlarge', 'r6in.32xlarge', 'r6in.4xlarge', 'r6in.8xlarge', 'r6in.large', 'r6in.metal', 'r6in.xlarge', 'r7a.12xlarge', 'r7a.16xlarge', 'r7a.24xlarge', 'r7a.2xlarge', 'r7a.32xlarge', 'r7a.48xlarge', 'r7a.4xlarge', 'r7a.8xlarge', 'r7a.large', 'r7a.medium', 'r7a.metal-48xl', 'r7a.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7gd.12xlarge', 'r7gd.16xlarge', 'r7gd.2xlarge', 'r7gd.4xlarge', 'r7gd.8xlarge', 'r7gd.large', 'r7gd.medium', 'r7gd.metal', 'r7gd.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 'r7iz.12xlarge', 'r7iz.16xlarge', 'r7iz.2xlarge', 'r7iz.32xlarge', 'r7iz.4xlarge', 'r7iz.8xlarge', 'r7iz.large', 'r7iz.metal-16xl', 'r7iz.metal-32xl', 'r7iz.xlarge', 'r8g.12xlarge', 'r8g.16xlarge', 'r8g.24xlarge', 'r8g.2xlarge', 'r8g.48xlarge', 'r8g.4xlarge', 'r8g.8xlarge', 'r8g.large', 'r8g.medium', 'r8g.metal-24xl', 'r8g.metal-48xl', 'r8g.xlarge', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-18tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge', 'x8g.12xlarge', 'x8g.16xlarge', 'x8g.24xlarge', 'x8g.2xlarge', 'x8g.48xlarge', 'x8g.4xlarge', 'x8g.8xlarge', 'x8g.large', 'x8g.medium', 'x8g.metal-24xl', 'x8g.metal-48xl', 'x8g.xlarge', 'z1d.12xlarge', 'z1d.2xlarge', 'z1d.3xlarge', 'z1d.6xlarge', 'z1d.large', 'z1d.metal', 'z1d.xlarge'],
              'sa-east-1': ['c1.medium', 'c1.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge', 'c3.large', 'c3.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge', 'c4.large', 'c4.xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.large', 'c5.metal', 'c5.xlarge', 'c5a.12xlarge', 'c5a.16xlarge', 'c5a.24xlarge', 'c5a.2xlarge', 'c5a.4xlarge', 'c5a.8xlarge', 'c5a.large', 'c5a.xlarge', 'c5ad.12xlarge', 'c5ad.16xlarge', 'c5ad.24xlarge', 'c5ad.2xlarge', 'c5ad.4xlarge', 'c5ad.8xlarge', 'c5ad.large', 'c5ad.xlarge', 'c5d.12xlarge', 'c5d.18xlarge', 'c5d.24xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.large', 'c5d.metal', 'c5d.xlarge', 'c5n.18xlarge', 'c5n.2xlarge', 'c5n.4xlarge', 'c5n.9xlarge', 'c5n.large', 'c5n.metal', 'c5n.xlarge', 'c6a.12xlarge', 'c6a.16xlarge', 'c6a.24xlarge', 'c6a.2xlarge', 'c6a.32xlarge', 'c6a.48xlarge', 'c6a.4xlarge', 'c6a.8xlarge', 'c6a.large', 'c6a.metal', 'c6a.xlarge', 'c6g.12xlarge', 'c6g.16xlarge', 'c6g.2xlarge', 'c6g.4xlarge', 'c6g.8xlarge', 'c6g.large', 'c6g.medium', 'c6g.metal', 'c6g.xlarge', 'c6gd.12xlarge', 'c6gd.16xlarge', 'c6gd.2xlarge', 'c6gd.4xlarge', 'c6gd.8xlarge', 'c6gd.large', 'c6gd.medium', 'c6gd.metal', 'c6gd.xlarge', 'c6gn.12xlarge', 'c6gn.16xlarge', 'c6gn.2xlarge', 'c6gn.4xlarge', 'c6gn.8xlarge', 'c6gn.large', 'c6gn.medium', 'c6gn.xlarge', 'c6i.12xlarge', 'c6i.16xlarge', 'c6i.24xlarge', 'c6i.2xlarge', 'c6i.32xlarge', 'c6i.4xlarge', 'c6i.8xlarge', 'c6i.large', 'c6i.metal', 'c6i.xlarge', 'c6id.12xlarge', 'c6id.16xlarge', 'c6id.24xlarge', 'c6id.2xlarge', 'c6id.32xlarge', 'c6id.4xlarge', 'c6id.8xlarge', 'c6id.large', 'c6id.metal', 'c6id.xlarge', 'c6in.12xlarge', 'c6in.16xlarge', 'c6in.24xlarge', 'c6in.2xlarge', 'c6in.32xlarge', 'c6in.4xlarge', 'c6in.8xlarge', 'c6in.large', 'c6in.metal', 'c6in.xlarge', 'c7g.12xlarge', 'c7g.16xlarge', 'c7g.2xlarge', 'c7g.4xlarge', 'c7g.8xlarge', 'c7g.large', 'c7g.medium', 'c7g.metal', 'c7g.xlarge', 'c7i-flex.12xlarge', 'c7i-flex.16xlarge', 'c7i-flex.2xlarge', 'c7i-flex.4xlarge', 'c7i-flex.8xlarge', 'c7i-flex.large', 'c7i-flex.xlarge', 'c7i.12xlarge', 'c7i.16xlarge', 'c7i.24xlarge', 'c7i.2xlarge', 'c7i.48xlarge', 'c7i.4xlarge', 'c7i.8xlarge', 'c7i.large', 'c7i.metal-24xl', 'c7i.metal-48xl', 'c7i.xlarge', 'g4dn.12xlarge', 'g4dn.16xlarge', 'g4dn.2xlarge', 'g4dn.4xlarge', 'g4dn.8xlarge', 'g4dn.metal', 'g4dn.xlarge', 'g5.12xlarge', 'g5.16xlarge', 'g5.24xlarge', 'g5.2xlarge', 'g5.48xlarge', 'g5.4xlarge', 'g5.8xlarge', 'g5.xlarge', 'g6.12xlarge', 'g6.16xlarge', 'g6.24xlarge', 'g6.2xlarge', 'g6.48xlarge', 'g6.4xlarge', 'g6.8xlarge', 'g6.xlarge', 'gr6.4xlarge', 'gr6.8xlarge', 'i3.16xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.large', 'i3.metal', 'i3.xlarge', 'i3en.12xlarge', 'i3en.24xlarge', 'i3en.2xlarge', 'i3en.3xlarge', 'i3en.6xlarge', 'i3en.large', 'i3en.metal', 'i3en.xlarge', 'i4i.12xlarge', 'i4i.16xlarge', 'i4i.24xlarge', 'i4i.2xlarge', 'i4i.32xlarge', 'i4i.4xlarge', 'i4i.8xlarge', 'i4i.large', 'i4i.metal', 'i4i.xlarge', 'inf1.24xlarge', 'inf1.2xlarge', 'inf1.6xlarge', 'inf1.xlarge', 'inf2.24xlarge', 'inf2.48xlarge', 'inf2.8xlarge', 'inf2.xlarge', 'm1.large', 'm1.medium', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge', 'm2.xlarge', 'm3.2xlarge', 'm3.large', 'm3.medium', 'm3.xlarge', 'm4.10xlarge', 'm4.16xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.large', 'm4.xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.large', 'm5.metal', 'm5.xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.large', 'm5a.xlarge', 'm5ad.12xlarge', 'm5ad.16xlarge', 'm5ad.24xlarge', 'm5ad.2xlarge', 'm5ad.4xlarge', 'm5ad.8xlarge', 'm5ad.large', 'm5ad.xlarge', 'm5d.12xlarge', 'm5d.16xlarge', 'm5d.24xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.8xlarge', 'm5d.large', 'm5d.metal', 'm5d.xlarge', 'm5zn.12xlarge', 'm5zn.2xlarge', 'm5zn.3xlarge', 'm5zn.6xlarge', 'm5zn.large', 'm5zn.metal', 'm5zn.xlarge', 'm6a.12xlarge', 'm6a.16xlarge', 'm6a.24xlarge', 'm6a.2xlarge', 'm6a.32xlarge', 'm6a.48xlarge', 'm6a.4xlarge', 'm6a.8xlarge', 'm6a.large', 'm6a.metal', 'm6a.xlarge', 'm6g.12xlarge', 'm6g.16xlarge', 'm6g.2xlarge', 'm6g.4xlarge', 'm6g.8xlarge', 'm6g.large', 'm6g.medium', 'm6g.metal', 'm6g.xlarge', 'm6gd.12xlarge', 'm6gd.16xlarge', 'm6gd.2xlarge', 'm6gd.4xlarge', 'm6gd.8xlarge', 'm6gd.large', 'm6gd.medium', 'm6gd.metal', 'm6gd.xlarge', 'm6i.12xlarge', 'm6i.16xlarge', 'm6i.24xlarge', 'm6i.2xlarge', 'm6i.32xlarge', 'm6i.4xlarge', 'm6i.8xlarge', 'm6i.large', 'm6i.metal', 'm6i.xlarge', 'm6id.12xlarge', 'm6id.16xlarge', 'm6id.24xlarge', 'm6id.2xlarge', 'm6id.32xlarge', 'm6id.4xlarge', 'm6id.8xlarge', 'm6id.large', 'm6id.metal', 'm6id.xlarge', 'm7g.12xlarge', 'm7g.16xlarge', 'm7g.2xlarge', 'm7g.4xlarge', 'm7g.8xlarge', 'm7g.large', 'm7g.medium', 'm7g.metal', 'm7g.xlarge', 'm7gd.12xlarge', 'm7gd.16xlarge', 'm7gd.2xlarge', 'm7gd.4xlarge', 'm7gd.8xlarge', 'm7gd.large', 'm7gd.medium', 'm7gd.metal', 'm7gd.xlarge', 'm7i-flex.12xlarge', 'm7i-flex.16xlarge', 'm7i-flex.2xlarge', 'm7i-flex.4xlarge', 'm7i-flex.8xlarge', 'm7i-flex.large', 'm7i-flex.xlarge', 'm7i.12xlarge', 'm7i.16xlarge', 'm7i.24xlarge', 'm7i.2xlarge', 'm7i.48xlarge', 'm7i.4xlarge', 'm7i.8xlarge', 'm7i.large', 'm7i.metal-24xl', 'm7i.metal-48xl', 'm7i.xlarge', 'p4d.24xlarge', 'p5.48xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge', 'r3.large', 'r3.xlarge', 'r4.16xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.large', 'r4.xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.large', 'r5.metal', 'r5.xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.large', 'r5a.xlarge', 'r5ad.12xlarge', 'r5ad.16xlarge', 'r5ad.24xlarge', 'r5ad.2xlarge', 'r5ad.4xlarge', 'r5ad.8xlarge', 'r5ad.large', 'r5ad.xlarge', 'r5b.12xlarge', 'r5b.16xlarge', 'r5b.24xlarge', 'r5b.2xlarge', 'r5b.4xlarge', 'r5b.8xlarge', 'r5b.large', 'r5b.metal', 'r5b.xlarge', 'r5d.12xlarge', 'r5d.16xlarge', 'r5d.24xlarge', 'r5d.2xlarge', 'r5d.4xlarge', 'r5d.8xlarge', 'r5d.large', 'r5d.metal', 'r5d.xlarge', 'r5n.12xlarge', 'r5n.16xlarge', 'r5n.24xlarge', 'r5n.2xlarge', 'r5n.4xlarge', 'r5n.8xlarge', 'r5n.large', 'r5n.metal', 'r5n.xlarge', 'r6g.12xlarge', 'r6g.16xlarge', 'r6g.2xlarge', 'r6g.4xlarge', 'r6g.8xlarge', 'r6g.large', 'r6g.medium', 'r6g.metal', 'r6g.xlarge', 'r6gd.12xlarge', 'r6gd.16xlarge', 'r6gd.2xlarge', 'r6gd.4xlarge', 'r6gd.8xlarge', 'r6gd.large', 'r6gd.medium', 'r6gd.metal', 'r6gd.xlarge', 'r6i.12xlarge', 'r6i.16xlarge', 'r6i.24xlarge', 'r6i.2xlarge', 'r6i.32xlarge', 'r6i.4xlarge', 'r6i.8xlarge', 'r6i.large', 'r6i.metal', 'r6i.xlarge', 'r7g.12xlarge', 'r7g.16xlarge', 'r7g.2xlarge', 'r7g.4xlarge', 'r7g.8xlarge', 'r7g.large', 'r7g.medium', 'r7g.metal', 'r7g.xlarge', 'r7i.12xlarge', 'r7i.16xlarge', 'r7i.24xlarge', 'r7i.2xlarge', 'r7i.48xlarge', 'r7i.4xlarge', 'r7i.8xlarge', 'r7i.large', 'r7i.metal-24xl', 'r7i.metal-48xl', 'r7i.xlarge', 't1.micro', 't2.2xlarge', 't2.large', 't2.medium', 't2.micro', 't2.nano', 't2.small', 't2.xlarge', 't3.2xlarge', 't3.large', 't3.medium', 't3.micro', 't3.nano', 't3.small', 't3.xlarge', 't3a.2xlarge', 't3a.large', 't3a.medium', 't3a.micro', 't3a.nano', 't3a.small', 't3a.xlarge', 't4g.2xlarge', 't4g.large', 't4g.medium', 't4g.micro', 't4g.nano', 't4g.small', 't4g.xlarge', 'u-12tb1.112xlarge', 'u-18tb1.112xlarge', 'u-3tb1.56xlarge', 'u-6tb1.112xlarge', 'u-6tb1.56xlarge', 'u-9tb1.112xlarge', 'x1.16xlarge', 'x1.32xlarge', 'x1e.16xlarge', 'x1e.2xlarge', 'x1e.32xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.xlarge', 'x2idn.16xlarge', 'x2idn.24xlarge', 'x2idn.32xlarge', 'x2idn.metal', 'x2iedn.16xlarge', 'x2iedn.24xlarge', 'x2iedn.2xlarge', 'x2iedn.32xlarge', 'x2iedn.4xlarge', 'x2iedn.8xlarge', 'x2iedn.metal', 'x2iedn.xlarge']}

ec2InstancesByRegion = {region: [E2Instances[i] for i in instances] for region, instances in regionDict.items()}
